/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    group.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    11 Dec 2000
 * Purpose:        Group class, equivalent to an HTML text block;
 *                 processing of tags;
 *                 processing of tables;
 *                 drawing of Groups.
 *----------------------------------------------------------------------
 * Changes:
 * 29 Nov 2000, zeyr@excite.com: color_lookup now used by FONT tag
 * 4 Dec 2000, zeyr@excite.com: drawing of Group background color
 * 5 Dec 2000, zeyr@excite.com: addition of strikethrough/overbar
 * 5 Dec 2000, zeyr@excite.com: cleaned up (some) debug printing
 * 7 Dec 2000, zeyr@excite.com: addition of MODEL tag for 3-D viewing
 * 8 Dec 2000, zeyr@excite.com: recommencing work on tables
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "defs.h"
#include "malloc.h"
#include "main.h"
#include "error.h"
#include "style.h"
#include "line.h"
#include "group.h"
#include "word.h"
#include "fonthash.h"
#include "table.h"
#include "color.h"
#include "3d.h"
#include "select.h"


enum { LIST_NONE=0, LIST_DOTTED=1, LIST_NUMBERED=2, LIST_DEFINITIONS=3 };

enum {
	TAG_NONE=0, TAG_P=1, TAG_BR=2,
	TAG_A, TAG_AUTHOR, TAG_ABBR, TAG_ACRONYM,
	TAG_B, TAG_BIG, TAG_BODY,
	TAG_CITE, TAG_CODE,
	TAG_DD, TAG_DFN, TAG_DL, TAG_DT, TAG_DATE, TAG_DEL,
	TAG_EM,
	TAG_FONT,
	TAG_H1, TAG_H2, TAG_H3, TAG_H4, TAG_H5, TAG_H6,
	TAG_HR, TAG_HTML, TAG_HEAD,
	TAG_I, TAG_IMG, TAG_INS,
	TAG_KBD,
	TAG_LI,
	TAG_NOSCRIPT,
	TAG_OL, TAG_OVER,
	TAG_PRE, 
	TAG_Q, 
	TAG_SMALL, TAG_SPAN, TAG_SCRIPT, TAG_SAMP, TAG_STRONG, TAG_STRIKE,
	TAG_SUP, TAG_SUB,
	TAG_TABLE, TAG_TD, TAG_TR, TAG_TH, TAG_TITLE, TAG_TBODY, TAG_THEAD,
	TAG_TT, 
	TAG_U, TAG_UL,
	TAG_VAR,

	/* Tags used for two-dimensional drawn images */
	TAG_DRAW, TAG_LINE, TAG_HLINE, TAG_VLINE, TAG_RECT, TAG_FILLEDRECT, 
	TAG_OVAL, TAG_FILLEDOVAL, TAG_TEXT, TAG_PIXEL, TAG_LINEREL,

	/* Tags used for three-dimensional interactive modeled forms */
	TAG_MODEL,

	TAG_MAXVALUE
};


/*-----------------------------------------------------------------------*/

static int
group_geometry_core (DisplayDevice *, Group*, Word *, int);

/*-----------------------Utility Routines--------------------------------*/


int 
my_min (int a, int b)
{
	return a<b?a:b;
}


int 
my_max (int a, int b)
{
	return a>b?a:b;
}


/*-----------------------Group Routines----------------------------------*/

Group *
group_new (DisplayDevice *dpy, Group *parent,
	Coordinate x0, Coordinate y0, 
	Dimension max_width, 
	Dimension max_height) 
{
	Group *g;
	Style *sl;

	g = (Group*) my_malloc(sizeof(Group));
	if(!g) 
		error_handler ("cannot allocate Group");

	if (max_width < 1 || max_height < 1)
		error_handler ("invalid coordinates for Group in group_new");

	memset ((void*)g, 0, sizeof(Group));

#if 0
	if(parent) {
		if (x0 < parent->x0) x0=parent->x0;
		if (y0 < parent->y0) y0=parent->y0;
		if (x1 > parent->x1) x1=parent->x1;
		if (y1 > parent->y1) y1=parent->y1;
	}
#endif

	/* Initialize Word-append tracking */
	g->x = 0; 
	g->xmax = 0;
	g->ymax = 0;

	/* Set up Group position and dimensions */
	g->x0=x0; 
	g->y0=y0; 
	g->max_width=max_width;
	g->max_height=max_height;

	/* These are relative to the top-left corner of the Group */
	g->xstart=0;
	g->ystart=0;

	/* These are content dimensions */
	g->width=0;
	g->height=0;

	g->bgcolor=INVALID_COLOR;
	g->border_color=INVALID_COLOR;
	g->border_size=0;

	g->parent=NULL;
	g->child=NULL;

	g->line_list=NULL;
	g->last_line=NULL;
	g->total_lines=0;

	g->list_type = LIST_NONE;
	g->list_indent = dpy->list_indentation_space;

	/* We really can't work without at least one Style.
	 */
	g->style_has_changed=TRUE;
	g->style_stored_to_word=FALSE;
	sl = style_new ();
	sl->family = my_strdup (dpy->family);
	sl->size = dpy->fontsize;
	g->style_stack = sl;
	style_fix_font (dpy, sl);

	return g;
}


Group*
group_parent (Group *g)
{
	CHECK_PARAM_NOT_NULL(g);
	return g->parent;
}


int
group_get_x (Group *g) {
	if (!g)
	  error_handler ("attempt to get group info when no groups exist");

	return g->x;
}

int
group_get_y (Group *g) {
	if (!g)
	  error_handler ("attempt to get group info when no groups exist");

	return g->ystart;
}



/* Routine to add an inter-word space to a Line.
 */
void
group_add_whitespace (DisplayDevice *dpy, Group *g, int ascent)
{
	double tmp;

	/* Here I use the Golden Mean to produce the spacing
	 * between words. Why? Because it works best for attaining 
	 * readability while still having high word density. 
	 * The Ancient Greeks discovered it and used it to produce 
	 * works of art and architecture with proportions that
	 * intuitively seem perfect. It is calculated as follows:
	 *    m + m^2 = 1
	 * =>  m = approx. 0.6815
	 */
	tmp = 1.0 - 0.6815;
	tmp *= ascent;
	if ((g->x + tmp) >= g->max_width) {
		group_end_line (dpy, g);
	} else {
		g->x += tmp;
		if (g->x > g->width)
			g->width = g->x;
	}
}



void 
group_push_style(DisplayDevice *dpy, Group *g, Style *sl)
{
	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(g);
	CHECK_PARAM_NOT_NULL(sl);

	/* Note! We DO NOT want to "update" the font metrics info
	 * before we push a Style, since we must allow a Style
	 * to contain incomplete information e.g. lack a size value,
	 * and the style update routine would add in missing data
	 * such as font size.
	 */
	sl->next=g->style_stack;
	g->style_stack=sl;
	g->style_has_changed=TRUE;
	g->style_stored_to_word=FALSE;
}


void 
group_drop_style(Group *g) {
	Style *old;
	CHECK_PARAM_NOT_NULL(g);

	old=g->style_stack;
	if (!old)
		error_handler ("attempt to drop Style from Group when there are no Styles");
	g->style_stack = old->next;
	style_free(old);
	g->style_has_changed=TRUE;
	g->style_stored_to_word=FALSE;
}


/* Routine to get the current ascent, NOT the ascent
 * waiting to be applied to the next Word. Only if there
 * is no current ascent do we get the total style.
 */
Dimension
group_get_current_ascent (DisplayDevice *dpy, Group *g)
{
	Style sl;
	Font *f;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(g);

	f = g->total_style.fontdata;
	if (f)
		return f->ascent;
	else {
		Dimension a;
		group_get_total_style (dpy, g, &sl);
		a = sl.fontdata->ascent;
		if (sl.family) my_free (sl.family);
		if (sl.name) my_free (sl.name);
		return a;
	}
}


void
group_get_total_style (DisplayDevice *dpy, Group *g, Style *return_sl)
{
	unsigned short attr=0;
	char *family=NULL;
	Color bg=INVALID_COLOR;
	Color fg=INVALID_COLOR;
	Dimension spacing=0;
	Dimension s_amount=0;
	FontSize size=INVALID_FONTSIZE;
	Style *sl;
	char *name=NULL;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(g);
	CHECK_PARAM_NOT_NULL(return_sl);

	if (!g->style_has_changed) {
		style_copy (return_sl, &g->total_style);
		return;
	}

	style_reset (return_sl);

	/* Get bitwise OR of style flags, plus get colors and fontsize,
	 * and font family.
	 */
	sl=g->style_stack;
	while(sl) {
		attr |= sl->flags;

		if (bg==INVALID_COLOR && sl->bg != INVALID_COLOR) 
			bg=sl->bg;

		if (fg==INVALID_COLOR && sl->fg != INVALID_COLOR) 
			fg=sl->fg;

		if (size==INVALID_FONTSIZE && 
		  sl->size != INVALID_FONTSIZE) {
			size=sl->size;
		}

		if (spacing==INVALID_DIMENSION && 
		  sl->interchar_spacing != INVALID_DIMENSION) 
			spacing = sl->interchar_spacing;

		if (s_amount==INVALID_DIMENSION && 
		  sl->super_amount != INVALID_DIMENSION) 
			s_amount = sl->super_amount;

		if (!family && sl->family)  {
			family=my_strdup(sl->family);
		}

		if (sl->name) {
			name=my_strdup(sl->name);
			break;
		}

		if (sl->flags & ATTR_PLAIN) 
			break;

		sl=sl->next;
	}

	/* Remove the plain marker. */
	attr &= (0xffff ^ ATTR_PLAIN);

	/* If per chance there is no value for any of these,
	 * use the default.
	 */
	return_sl->name = name;
	return_sl->flags = attr;
	return_sl->size = size;
	return_sl->family = family;
	return_sl->interchar_spacing = spacing!=INVALID_DIMENSION ? spacing : 0;
	return_sl->super_amount = s_amount!=INVALID_DIMENSION ? s_amount : 0;
	return_sl->fg = fg!=INVALID_COLOR ? fg : dpy->fgcolor;
	return_sl->bg = bg!=INVALID_COLOR ? bg : dpy->bgcolor;
	return_sl->fontdata = fonthash_lookup (dpy, family, size, attr);

	style_copy (&g->total_style, return_sl);
	g->style_has_changed=FALSE;
	g->style_stored_to_word=FALSE;

#if 0
printf ("returning total style: ");
style_dump (logfile, &g->total_style);
printf ("\n");
fflush(stdout);
#endif
}


int
group_get_fontsize (Group *g)
{
	FontSize size=INVALID_FONTSIZE;
	Style *sl;

	CHECK_PARAM_NOT_NULL(g);

	/* Now get colors and fontsize */
	sl=g->style_stack;
	while(sl) {
		if (size==INVALID_FONTSIZE && sl->size!=INVALID_FONTSIZE) {
			size=sl->size;
			break;
		}
		sl=sl->next;
	}

	return size;
}


void 
group_free (Group *g) {
	Style *sl;

	CHECK_PARAM_NOT_NULL(g);

	sl=g->style_stack;
	while (sl) {
		Style *next;
		next = sl->next;
		style_free(sl);
		sl = next;
	}
	my_free((char*) g);

	/* Don't remove the Words themselves,
	 * since table_geometry could be calling
	 * this routine. XX - need a "light"
 	 * version of this routine.
	 */
	if (g->line_list) my_free ((char*) g->line_list);
	if (g->work_line) my_free ((char*) g->work_line);
}



void
group_make_spacer_line (Group *g, Dimension vertical_size)
{
	CHECK_PARAM_NOT_NULL(g);

	ENTER;

/* printf("REACHED: group_make_spacer_line\n"); */

	/* Create a new spacer line and update everything.
	 * Reuse an existing empty line.
	 */
	if (!g->work_line || (!g->work_line->ascent))
		group_make_line (g);

	g->work_line->ascent = vertical_size;
	g->work_line->descent = 0;
}


void
group_make_line (Group *g)
{
	CHECK_PARAM_NOT_NULL(g);

	ENTER;

	if (g->work_line)  {
		if (g->work_line->descent || g->work_line->ascent)
			error_handler ("cannot create new work line because there already is one");

		/* If we reach this point, there was an empty work line
		 * and we can save time by reusing it.
		 */
	}
	else 
		g->work_line = line_new();

	g->work_line->next = NULL;
}



void
group_append_word (DisplayDevice *dpy, Group *g, Word *w) 
{
	Line *l;
	char *str;
	int tmp;

	CHECK_PARAM_NOT_NULL(g);
	CHECK_PARAM_NOT_NULL(w);

	ENTER;

	str=w->str;

#if 0
	if (*str==' ') {
	fprintf (logfile, "FOUND SPACE when line="); 
	if(g->work_line) line_dump(logfile, g->work_line);
	else fprintf (logfile,"no work line\n");
	fflush(logfile);
	}
#endif

	if (debug_mode && logfile) {
		if (g->style_has_changed) fprintf (logfile,"APPEND: called with style-has-changed set\n");
		if (!g->style_stored_to_word) fprintf (logfile, "APPEND: Style not yet stored to word\n");
		fflush(logfile);
	}

	/* Ignore non-tag words with bad dimension values.
	 */
	if (*str!='<' && (w->width < 1 || (w->ascent+w->descent) < 1)) {
		fprintf (logfile, "APPEND: ignoring dimensionless non-tag Word %s\n",str);
		fflush(logfile);
		return;
	}

	if (w->min_width > g->min_width)
		g->min_width = w->min_width;

	/* If we are at the start of a line, and we are given a space
	 * char, then we can ignore it. 
	 * 
	 * It is important that we do this before looking at the
	 * paragraph space flag since honoring blanks would lead
	 * to consecutive paragraph spaces NOT being merged.
	 */
	if ((!g->work_line || !g->work_line->word_list) && ' '==str[0]) {
		return;
	}

	/* If a paragraph start has been flagged, add it now.
	 */
	if (g->insert_paragraph_space) {

#if 0
fprintf (logfile,"APPEND: inserting para space when str=%s\n", w->str);
if (g->work_line) {
fprintf (logfile, "and workline=\n"); line_dump(logfile, g->work_line);
fflush(logfile);
}
#endif
		group_end_line (dpy, g);
		group_make_spacer_line (g, 7 /*dpy->interparagraph_space*/);
		group_end_line (dpy, g);
		g->insert_paragraph_space=FALSE;
	}

	/* If we are at the start of a line, and we are given a space
	 * char, then we can ignore it.
	 *
	 * It is important that we do this AFTER looking at the
	 * paragraph space flag since honoring blanks after a paragraph
	 * space has been added could lead to stray whitespace Words
	 * being inserted at the starts of lines. 
	 */
	if ((!g->work_line || !g->work_line->word_list) && ' '==str[0]) {
		return;
	} else {
		if (*str==' ') {
#if 0
fprintf (logfile, "APPEND: we have a space word\n");
if (g->work_line) line_dump (logfile, g->work_line); else fprintf (logfile, "no work line\n");
fflush(logfile);
#endif
		}
	}

	/* Set the coordinates for the Word, performing a line wrap
	 * if necessary.
	 */
	tmp = g->x + w->width;
	if (tmp > g->max_width) {
		group_end_line_make_line (dpy, g);
		w->x = 0;
		tmp = g->x + w->width;
	} else {
		w->x = g->x;
	}
	w->y = g->ystart;

	if (tmp > g->width)
		g->width = tmp;

#if 0
fprintf (logfile,"APPEND: placing word %s at %ld %ld (g->x=%lu, g->width=%lu)\n", w->str, w->x, w->y, g->x,g->width);
fprintf (logfile,"group info: x0,y0=%ld,%ld xstart,ystart=%ld,%ld\n",
g->x0,g->y0,g->xstart,g->ystart);
fflush(logfile);
#endif

	if (!g->work_line) 
		group_make_line (g);

	l = g->work_line;

	/* If the current and previous words are both blank spaces,
	 * then skip the current one.
	 */
	if (' '==str[0] && l->word_list) {
		if (' '==l->last_word->str[0]) {
			/* Adjacent spaces are ignored. */
			return;
		}
	}

	/* Create a new Style based on the total current style data,
	 * if and only if the cumulative style has recently been
	 * changed.
	 */
	if (' '!=str[0]) {
		if (g->style_has_changed || !g->style_stored_to_word)  {
			w->style = style_new();
			group_get_total_style (dpy, g, w->style);
#if 0
// printf ("top of style stack: ");
// style_dump(logfile, g->style_stack);
// printf ("\n");
#endif
			g->style_stored_to_word=TRUE;
		}
		else
			w->style = NULL;

		if (debug_mode && logfile) {
			fprintf(logfile, "APPEND: calling add_word\n");
			fflush(logfile);
		}
		line_add_word (l, w);
	}

	g->x += w->width;

	/* Transfer flags to allow later output of HTML */
	if (g->beginning_new_line) {
		w->flags |= WORD_HAS_PRECEDING_LINE_BREAK;
		g->beginning_new_line=FALSE;
	}
	if (g->beginning_new_paragraph) {
		w->flags |= WORD_HAS_PRECEDING_PARAGRAPH_BREAK;
		w->para_align = g->new_paragraph_align;
		g->beginning_new_paragraph=FALSE;
	}

#if 0
	if (g->beginning_heading1) {
		w->flags |= WORD_HAS_PRECEDING_HEADING1;
		g->beginning_heading1=FALSE;
	}
	if (g->beginning_heading2) {
		w->flags |= WORD_HAS_PRECEDING_HEADING2;
		g->beginning_heading2=FALSE;
	}
	if (g->beginning_heading3) {
		w->flags |= WORD_HAS_PRECEDING_HEADING3;
		g->beginning_heading3=FALSE;
	}
	if (g->beginning_heading4) {
		w->flags |= WORD_HAS_PRECEDING_HEADING4;
		g->beginning_heading4=FALSE;
	}
	if (g->beginning_heading5) {
		w->flags |= WORD_HAS_PRECEDING_HEADING5;
		g->beginning_heading5=FALSE;
	}
	if (g->beginning_heading6) {
		w->flags |= WORD_HAS_PRECEDING_HEADING6;
		g->beginning_heading6=FALSE;
	}
#endif

	if (debug_mode && logfile) {
		if (str[0]!=' ') {
			if (w->style) {
				fprintf (logfile,"APPEND: added word %s at %ld %ld, (%lu by [%lu %lu])", w->str, w->x,w->y,w->width,w->ascent,w->descent);
				style_dump (logfile, w->style);
				fprintf (logfile,"\n");
			}
			else
				fprintf (logfile,"APPEND: added word %s at %ld %ld (%lu by [%lu %lu]), no style\n", w->str,w->x,w->y,w->width, w->ascent,w->descent);
		}
		else
			fprintf (logfile, "APPEND: did not add: '%s'\n", w->str);

		fflush(logfile);
	}

	/* Now, if the parser decided to save some memory by setting
	 * the "following whitespace" flag rather than allocate
	 * ~50 bytes for a whitespace-containing Word, we in response must
	 * update things to reflect that.
	 */
	if (w->flags & WORD_HAS_FOLLOWING_WHITESPACE) {
		group_add_whitespace (dpy, g, w->ascent);
	}
#if 0
fprintf(logfile,"-------------finished append of %s, width is %lu \n", w->str,g->width);
fflush(logfile);
#endif

	/* If we encounter a tag which can itself lead to a change
	 * in the current Style stored in the display interface,
	 * then let's tell ourselves to update it with the next word.
	 */
	if (w->tag_number == TAG_TABLE) {
		g->style_stored_to_word = FALSE;
	}

	if (g->x == g->max_width)
		group_end_line (dpy, g);
}


/* Update y values for objects on line in order to perform y finishment.
 */
void
group_end_line (DisplayDevice *dpy, Group *g)
{
	Line *line;
	Word *w;

	CHECK_PARAM_NOT_NULL(g);

	ENTER;

	line = g->work_line;

	/* If there is no work Line, just return. Example scenario: <H1>
	 * occuring as the first word of the body. We'd try to terminate
	 * the current line, even though there isn't one.
	 */
	if (!line) {
		if (debug_mode && logfile)
			fprintf (logfile, "GROUP_END_LINE: no work line\n");
		return;
	}

	/* If there is simply no data on this line (height==0)
	 * then don't add it, but don't remove it either.
	 */
#if 0
	if (!line->ascent && !line->descent) {
		if (debug_mode && logfile)
		fprintf (logfile, "LINE HAS NO HEIGHT\n");
		return;
	}
#endif

	/* If we have words, finish them.
	 */
	w=line->word_list;
	if (w) {
		if (debug_mode && logfile)
			line_dump (logfile, line);

		if (debug_mode && logfile) {
			fprintf (logfile, "ALIGN: first word = %s\n",w?(char*)w->str:"(no word)");
			fflush(logfile);
		}

		while(w)
		{ 
			int adjust = line->ascent - w->ascent ; 
if (!line->ascent)
 warning_handler ("uh oh, line has no ascent");

			if (debug_mode && logfile) {
				fprintf (logfile, "ALIGN: line height=[%lu+%lu], obj height=[%lu+%lu]\n", 
				  line->ascent, line->descent,
				   w->ascent, w->descent);
			}

			if (adjust<0)
				error_handler ("negative vertical adjust value in finalize_line");

			if (debug_mode && logfile) {
				fprintf (logfile, "ALIGN: object '%s' was at here: x,y,wid,asc,desc of %ld,%ld,%ld [%lu+%lu]\n",
				  w->str,w->x,w->y,w->width,w->ascent,w->descent);
			}

			/* XX recursive better? */
			w->y += adjust; 

#if 0
			if (w->subgroup) {
				Group *subgroup = (Group*) w->subgroup;
				subgroup->y0 += adjust;
			}
#endif

			if (debug_mode && logfile) {
				fprintf (logfile, "ALIGN: object '%s' moved to x,y,w,asc,desc of %ld,%ld,%ld,%ld,%ld\n",
					w->str,
					w->x,w->y,
					w->width,w->ascent,w->descent);
				fflush(logfile);
			}

			w=w->next_in_line;
		}
	}

	/* Position the line, even if empty! */
	line->y = g->ystart;

	g->xstart = 0; /* Notice, (XXX) no indentation value used... */

#if 0
	line = g->work_line;
	if (!line) 
		return;
#endif

/* printf (" firstword=%s\n", line->word_list?line->word_list->str:"(none)"); */

	++g->total_lines;

	g->ystart += line->ascent + line->descent;
	g->height += line->ascent + line->descent;

	g->ystart += dpy->interline_padding;
	g->height += dpy->interline_padding;

	g->xstart = 0;
	g->x = 0; 


#if 0
	if (g->width < line->width) 
		g->width = line->width; 
#endif

	/* Move the Line over into the Line list */
	if(!g->line_list) {
		g->line_list = line;
		g->last_line = line;
	} else {
		g->last_line->next = line;
		g->last_line = line;
	}
	g->work_line->next = NULL;
	g->work_line = NULL;

	g->style_has_changed=TRUE;
	g->style_stored_to_word= FALSE;
}


/* Routine to finish, then terminate a line, creating a new one.
 *
 * Things to do when terminating a line:
 * 0. Save the line height value.
 * 1. Align the line (clears line height since work_line becomes NULL).
 * 2. Increase the ystart value by the height value.
 * 
 */
void
group_end_line_make_line (DisplayDevice *dpy, Group *g)
{
	int line_height;

	CHECK_PARAM_NOT_NULL(g);	

	/* If there is no work line, don't despair, just skip that part.
	 */
	if (g->work_line) {
		line_height = g->work_line->ascent + g->work_line->descent;
		group_end_line (dpy, g);

		/* If end_line didn't touch the work line, that means it
		 * had no height. We will now preserve it rather than
		 * remove it and reallocate a new one.
		 */
		if (g->work_line)
			return;
	}

	group_make_line(g);
}


/* Use current total style info to get text dimensions.
 */
void
group_text_dims (DisplayDevice *dpy, Group *g, char *text, 
  Dimension *wid, Dimension *ascent, Dimension *descent)
{
	Style sl;

	CHECK_PARAM_NOT_NULL(dpy);	
	CHECK_PARAM_NOT_NULL(g);	
	CHECK_PARAM_NOT_NULL(text);	
	CHECK_PARAM_NOT_NULL(wid);	
	CHECK_PARAM_NOT_NULL(ascent);	
	CHECK_PARAM_NOT_NULL(descent);	

	group_get_total_style (dpy, g, &sl);

#if 0
fprintf (logfile, "GROUPTEXTDIMS: just after group_get_total_style, in text_dims, we have: ");
style_dump(logfile, &sl);
fprintf (logfile,"\n  ^--- that's for '%s'\n", text);
fflush(logfile);
#endif

	dpy->text_dims (text, &sl, wid, ascent, descent);

	/* If the font information was incorrect, we will need to copy 
	 * over the new font info to improve our efficiency. Otherwise
	 * each call to text_dims would result in one or more calls
	 * to fonthash_lookup, as would calls to dpy->draw_text.
	 */
	if (g->total_style.fontdata != sl.fontdata)
		style_copy (&g->total_style, &sl);
}


void
group_redraw_core (DisplayDevice *dpy, Group *g, Coordinate rx, Coordinate ry,
	Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1,
	Coordinate parents_x, Coordinate parents_y)
{
	char *str;
	Line *l;
	Coordinate u_x0, u_x1, u_y;
	Coordinate o_x0, o_x1, o_y;
	Coordinate s_x0, s_x1, s_y;
	Coordinate start_y, end_y;
	Coordinate base_x, base_y;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(g);

	/* Avoid unnecessary drawing */
	dpy->set_clip (x0,y0,x1,y1);

	start_y = ry + y0;
	end_y = ry + y1;

	/* Reset underline tracking */
	u_x0 = u_x1 = u_y = INVALID_COORDINATE;

	/* Reset overline tracking */
	o_x0 = o_x1 = o_y = INVALID_COORDINATE;

	/* Reset strikethrough tracking */
	s_x0 = s_x1 = s_y = INVALID_COORDINATE;

	if (debug_mode && logfile) {
		fprintf (logfile, "REDRAW: starty=%ld endy=%ld\n",start_y,end_y);
		fflush(logfile);
	}
 
	while (g) {
		Word *w;

		base_x = parents_x + g->x0;
		base_y = parents_y + g->y0;

		if (debug_mode && logfile) {
			int x0,y0,x1,y1;
			x0=base_x-rx;
			x1=base_x+g->width-rx;
			y0=base_y-ry;
			y1=base_y+g->height-ry;
			dpy->draw_rectangle (x0,y0,x1,y1);
		}

		if (g->bgcolor != INVALID_COLOR) {
			int x0,y0,x1,y1;

			dpy->set_colors (g->bgcolor,0);
			x0=base_x-rx;
			x1=base_x+g->width-rx;
			y0=base_y-ry;
			y1=base_y+g->height-ry;
			dpy->filled_rectangle (x0,y0,x1,y1);
			dpy->reset_colors ();
		}

		/* Initialize the LI tag counter for OL lists. */
		g->list_value = g->list_start;

		l = g->line_list;

		while (l) 
		{
			Coordinate ly0 = l->y;
			Coordinate ly1 = l->y + l->ascent + l->descent;

			/* The word_apply_offset call done in 
			 * process_table_tag does not of course affect Line
			 * information, so here we must adjust Line info
			 * on the fly.
			 */
			ly0 += base_y;
			ly1 += base_y;

			/* Draw any underlines that were started on a
			 * previous line.
			 */
			if (u_x0 != INVALID_COORDINATE) 
fprintf (logfile, "REDRAW: internal error: underline has been continued to next line\n");
			if (s_x0 != INVALID_COORDINATE) 
fprintf (logfile, "REDRAW: internal error: strike-through has been continued to next line\n");
			if (o_x0 != INVALID_COORDINATE) 
fprintf (logfile, "REDRAW: internal error: overline has been continued to next line\n");

			/* If we are within the area to be redrawn,
			 * then let us work on this line.
			 */
			if (ly0 <= end_y && ly1 >= start_y) {

				w = l->word_list; 

				while (w) {
					Style *sl = w->style;

					if (sl && w->tag_number == TAG_NONE) {

#if 0
						if (u_x0 !=INVALID_COORDINATE &&
						    !(sl->flags & ATTR_UNDERLINE)
							) 
						{
							u_x1 = w->x -1 ;
							dpy->draw_hline (
							  base_x+  u_x0-rx, 
							  base_x + u_x1-rx, 
							  base_y + u_y-ry);
							u_x0 = INVALID_COORDINATE;
							u_x1 = INVALID_COORDINATE;
							u_y  = INVALID_COORDINATE;
						}
						
						if (s_x0 !=INVALID_COORDINATE &&
						    !(sl->flags & ATTR_STRIKE)
							) 
						{
							s_x1 = w->x - 1;
							dpy->draw_hline (
							  base_x + s_x0-rx, 
							  base_x + s_x1-rx, 
							  base_y + s_y-ry);
							s_x0 = INVALID_COORDINATE;
							s_x1 = INVALID_COORDINATE;
							s_y  = INVALID_COORDINATE;
						}

						if (o_x0 !=INVALID_COORDINATE &&
						    !(sl->flags & ATTR_OVERBAR)
							) 
						{
							o_x1 = w->x - 1;
							dpy->draw_hline (
							  base_x + o_x0-rx, 
							  base_x + o_x1-rx, 
							  base_y + o_y-ry);
							o_x0 = INVALID_COORDINATE;
							o_x1 = INVALID_COORDINATE;
							o_y  = INVALID_COORDINATE;
						}
#endif

						if (u_x0==INVALID_COORDINATE  && 
						    (sl->flags & ATTR_UNDERLINE)) 
						{
							u_x0 = w->x;
							u_y = w->y + w->ascent + 1;
							u_x1 = INVALID_COORDINATE;
						}

						if (s_x0==INVALID_COORDINATE  && 
						    (sl->flags & ATTR_STRIKE)) 
						{
							s_x0 = w->x;
							s_y = w->y + w->ascent - w->ascent/3 - 1;
							s_x1 = INVALID_COORDINATE;
						}

						if (o_x0==INVALID_COORDINATE  && 
						    (sl->flags & ATTR_OVERBAR)) 
						{
							o_x0 = w->x;
							o_y = w->y;
							o_x1 = INVALID_COORDINATE;
						}
					}

#if 0
					if (w->str[0] != '<' &&
					    u_x0 !=INVALID_COORDINATE &&
					    w->next_in_line && 
					    w->next_in_line->style &&
					    !(w->next_in_line->style->flags & ATTR_UNDERLINE)
						) 
					{
						u_x1 = w->x + w->width - 1;
						dpy->draw_hline (
						  base_x + u_x0-rx, 
						  base_x + u_x1-rx, 
						  base_y + u_y-ry);
						u_x0 = INVALID_COORDINATE;
						u_x1 = INVALID_COORDINATE;
						u_y  = INVALID_COORDINATE;
					}

					if (w->str[0] != '<' &&
					    s_x0 !=INVALID_COORDINATE &&
					    w->next_in_line && 
					    w->next_in_line->style &&
					    !(w->next_in_line->style->flags & ATTR_STRIKE)
						) 
					{
						s_x1 = w->x + w->width - 1;
						dpy->draw_hline (
						  base_x + s_x0-rx, 
						  base_x + s_x1-rx, 
						  base_y + s_y-ry);
						s_x0 = INVALID_COORDINATE;
						s_x1 = INVALID_COORDINATE;
						s_y  = INVALID_COORDINATE;
					}
#endif
					
					str = w->str;
					if (!str)
						error_handler ("word lacks string");

					if (sl) {
						dpy->set_style (sl);
						if (debug_mode && logfile) 
						{
						  fprintf (logfile, "REDRAW: drawing Word %s which has this Style: ",
						    w->str);
						  fprintf (logfile, " size=%d\n", (int) sl->size);
						  fprintf (logfile, " attr=%d\n", sl->flags);
						  fprintf (logfile, " fg=%08lx\n", sl->fg);
						  fprintf (logfile, " bg=%08lx\n", sl->bg);
						  fflush(logfile);
						}
					}

					if ('&' == str[0]) {
						/* Assume the charset is ANSI.
						 */
						int ch=w->tag.datum;
						char str2[2];

						str2[0]=ch>=32 ? ch : '?';
						str2[1]=0;

						dpy->draw_text (
						  base_x + w->x-rx,
						  base_y + w->y-ry,
						  str2,
						  (w->flags & WORD_HAS_FOLLOWING_WHITESPACE) ? 1:0);

					} else if ('<' != str[0]) {
						/* An obj with negative x coordinate
						 * is supposed to be invisible.
						 */
						if (w->x != -1) {
							dpy->draw_text (
							  base_x + w->x-rx,
							  base_y + w->y-ry,
							  str,
							  (w->flags & WORD_HAS_FOLLOWING_WHITESPACE) ? 1:0);
							if (debug_mode && logfile) {
								fprintf (logfile,"REDRAW: drawing %s at %ld,%ld\n", str, w->x, w->y);
						  		fflush(logfile);
							}
						}
					} else {
						int num = w->tag_number;

						switch (num) {
						case TAG_TABLE: {
							Table *t = w->tag.table;
							if (t) {
								table_redraw (dpy, t,
								  rx,ry,x0,y0,x1,y1,
								  base_x,base_y);
							} else {
								error_handler ("table tag lacks Table object");
							}
							break;
						}
						case TAG_LI: {
							String str = (String) w->tag.ptr;
							Group *subgroup = (Group*) w->subgroup;

							if (*str=='&') {
								if (!strcmp("&SQUARE;",str) ||
								    !strcmp("&CIRCLE;",str) ||
								    !strcmp("&DISC;",str))
								{
									Coordinate cx,cy;
									cx = w->x + w->width/2;
									cy = w->y + w->ascent/2;
									dpy->reset_colors();
									dpy->draw_rectangle (
									  base_x + cx-3,
									  base_y + cy-3,
									  base_x + cx+3,
									  base_y + cy+3);
								} else {
									char str2[2];
									int ch;

									ch = dpy->convert_special_expr(str);
									str2[0]=ch>=32?ch:'?';
									str2[1]=0;

									dpy->draw_text (
									  base_x + w->x-rx,
									  base_y + w->y-ry,
									  str2,
									  (w->flags & WORD_HAS_FOLLOWING_WHITESPACE) ? 1:0);
								}
							} else {
									dpy->draw_text (
									  base_x + w->x-rx,
									  base_y + w->y-ry,
									  str,
									  (w->flags & WORD_HAS_FOLLOWING_WHITESPACE) ? 1:0);
							}
							group_redraw_core (dpy, subgroup, rx,ry,x0,y0,x1,y1,
								base_x, base_y);
							break;
						}
						case TAG_DRAW:
						case TAG_MODEL: /* for debug */
						case TAG_IMG: {
							Image *im;
							im = (Image*) w->tag.ptr;
							if (!im) {
								dpy->reset_colors();
								dpy->draw_x (
								  base_x + w->x-rx, 
								  base_y + w->y-ry,
								  base_x + w->x+w->width-1-rx, 
								  base_y + w->y+w->ascent-1-ry);
								dpy->draw_rectangle (
								  base_x + w->x-rx, 
								  base_y + w->y-ry,
								  base_x + w->x+w->width-1-rx, 
								  base_y + w->y+w->ascent-1-ry);
							} else {
								dpy->put_image (
								  base_x + w->x-rx, 
								  base_y + w->y-ry, 
								  im);
								if (debug_mode) {
									dpy->draw_rectangle (
									  base_x + w->x-rx, 
									  base_y + w->y-ry,
									  base_x + w->x+w->width-1-rx, 
									  base_y + w->y+w->ascent-1-ry);
								}
							}
							break;
						}

						case TAG_BODY: 
						case TAG_HTML: 
						case TAG_P: 
						case TAG_OL: 
						case TAG_UL: 
						/* case TAG_LI: */
						case TAG_DL: 
						case TAG_DT: 
						case TAG_DD: 
						case TAG_H1: 
						case TAG_H2: 
						case TAG_H3: 
						case TAG_H4: 
						case TAG_H5: 
						case TAG_H6:
						{
							Group *subgroup = (Group*) w->subgroup;
							if (subgroup)
							  group_redraw_core (dpy, subgroup, rx,ry,x0,y0,x1,y1,
								base_x, base_y);
							break;
						}

						case TAG_HR: {
							Dimension width = g->max_width-2;
							Dimension height = w->tag.datum;
							int i;
							Coordinate x,y;

							x = w->x-rx;
							y = w->y-ry+(w->ascent-height)/2;
							for (i=0; i<height; i++)
								dpy->draw_hline(base_x + x+1, 
								  base_x + x+width-1, 
								  base_y + y+i);
							break;
						}
						} /* switch */
					}

					/* Store an X coordinate in case we need to
					 * finish an underline.
					 */
					if (!w->next_in_line) {
						if (u_x0 != INVALID_COORDINATE)
							u_x1 = w->x + w->width;
						if (s_x0 != INVALID_COORDINATE)
							s_x1 = w->x + w->width;
						if (o_x0 != INVALID_COORDINATE)
							o_x1 = w->x + w->width;
					}

					w=w->next_in_line;
				}

#if 0
				if (u_x0 != INVALID_COORDINATE)
				{
					dpy->draw_hline (u_x0-rx, u_x1-rx, u_y-ry);
					u_x0 = INVALID_COORDINATE;
					u_x1 = INVALID_COORDINATE;
				}
				if (s_x0 != INVALID_COORDINATE)
				{
					dpy->draw_hline (s_x0-rx, s_x1-rx, s_y-ry);
					s_x0 = INVALID_COORDINATE;
					s_x1 = INVALID_COORDINATE;
				}
				if (o_x0 != INVALID_COORDINATE)
				{
					dpy->draw_hline (o_x0-rx, o_x1-rx, o_y-ry);
					o_x0 = INVALID_COORDINATE;
					o_x1 = INVALID_COORDINATE;
				}
#endif
			}

			/* Update any LI counter, regardless of whether
			 * this line will be drawn.
			 */
			w = l->word_list;
			if (w && w->tag_number==TAG_LI) {
				g->list_value += g->list_offset; 
			}

			l=l->next;
		}

		g=NULL;
	}

	dpy->reset_clip ();
	return;
}


/*------------------------------------------------------------------------
 * Name:     group_redraw
 * Purpose:  Goes through Group tree and draws each Word that needs
 *           to be drawn, IFF it has coordinates. A call to 
 *           group_geometry will assign needed coordinates.
 * Params:   rx,ry : the relative position of the top-left corner of
 *           the screen
 *           (x0,y0)-(x1,y1): area on-screen to draw
 *----------------------------------------------------------------------*/

void
group_redraw (DisplayDevice *dpy, Group *g, Coordinate rx, Coordinate ry,
	Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1)
{
	group_redraw_core (dpy,g,rx,ry,x0,y0,x1,y1,0,0);
}


void
group_dump (FILE*f, Group *g) {
#if 0
	while (g) 
	{

		fprintf (f, "GROUP name=%s max=%ld x %ld size=%ld x %ld  { ", 
			g->name?(char*)g->name:"(none)",
			g->max_width, g->max_height,
			g->width, g->height);
		fprintf (f, "lines=\n");
		if (!g->line_list) {
			fprintf (f, "-NONE-\n");
		} else {
			linelist_dump (logfile, g->line_list);
		}
		fflush(logfile);

		if (g->child) {
			fprintf (logfile, "sub");
			group_dump (logfile, g->child);
			fprintf (logfile, "\n");
		}
		fprintf (logfile, "\n}\n");
		fflush(logfile);
	}
#endif
}


/*========Table Geometry===============================================*/

int
table_geometry_single_row (DisplayDevice *dpy, Group *current_group,
  Table *tbl, Word *w, int row_number,
	Dimension *return_row_width,
	Dimension *return_row_height)
{
	int column_number;
	Dimension width_of_contents_only=0;
	Dimension height_of_contents_only=0;
	Group *subgroup;
	CellInfo *cells;

	ENTER;

	cells = &tbl->cells [tbl->columns * row_number];

	column_number=0;
	while(w) {
		if (w->tag_number == TAG_NONE) {
			w=w->next;
			continue;
		}

		if (w->tag_number == TAG_TD) {
			CellInfo *cell;
			char *str = word_fetch_tag_var_value (w,"width");
			char *str2 = word_fetch_tag_var_value (w,"colspan");
			char *str3 = word_fetch_tag_var_value (w,"rowspan");
			char *str4 = word_fetch_tag_var_value (w,"height");

			do {
				cell = &cells[column_number];
				if (!cell->is_leading_cell) {
					cell = NULL;
					column_number++;
					if (column_number >= tbl->columns)
						error_handler ("colspan tracking is wrong");
				}
			}
			while (!cell);

			/* Store any width requests.
			 */
			if (str && isdigit (*str)) {
				long value = atoi(str);
				int is_per = strchr(str,'%') ? TRUE : FALSE;
				if (is_per && value>0) {
					Dimension wid = tbl->max_content_width;
					wid *= 100;
					wid /= value;
					cell->desired_width = wid;
				  
				} else {
					if (value < 0) value=0;
					cell->desired_width = value;
				}
			}

			/* Store any height requests.
			 *
			 * Note, the height=X% parameter is kind of 
			 * meaningless but I'm making it to be simply
			 * percentage-of-screen-height.
			 */
			if (str && isdigit (*str)) {
				long value = atoi(str);
				int is_per = strchr(str,'%') ? TRUE : FALSE;
				if (is_per && value>0) {
					Dimension ht = dpy->height;
					ht *= 100;
					ht /= value;
					cell->desired_height = ht;
				  
				} else {
					if (value < 0) value=0;
					cell->desired_height = value;
				}
			}

			/* Column span request */
			if (str2) {
				int value=atoi (str2);
				if (value >= 1) 
					cell->col_span = value;
			} 

			/* Row span request */
			if (str3) {
				int value=atoi (str3);
				if (value >= 1) 
					cell->row_span = value;
			}

			/* Clear the is_leading_cell flags as
			 * necessary for cells which have rowspan
			 * and colspan values > 1.
			 */
			if (cell->col_span > 1 || cell->row_span > 1) {
				int i,j;
				CellInfo *c;

				for (i=0; i<cell->col_span; i++) {
					for (j=0; j<cell->row_span; j++) {
						int ix =tbl->columns*(j+row_number)+column_number+i;
						c = &tbl->cells[ix];
						if (i || j) 
							c->is_leading_cell=FALSE;
						else
							c->is_leading_cell=TRUE;
					}
				}
			}

			/* Get the optimal and minimum width for this column.
			 */
			if (w->child) 
			{
				subgroup=group_new (dpy, NULL, 
				  0,0,current_group->max_width,MAX_COORD);

				/* Get cell dimensions (best case and minimal)
				 */
				group_geometry (dpy, subgroup, w->child);

				/* Store the optimal and minimum widths and
				 * the height. 
				 */
				cell->optimal_width = subgroup->width;
				cell->minimal_width = subgroup->min_width;
				cell->optimal_height = subgroup->height;

				/* Store the pointer to the <TD> Word.
				 */
				cell->td = w;

				/* Update our overall row width, vital for single-
				 * column tables.
				 */
				width_of_contents_only += subgroup->width;

				/* Update our row height.
				 */
				if (subgroup->height > height_of_contents_only)
				  height_of_contents_only = subgroup->height;

#if 0
fprintf (logfile,"TABLE-GEO-SINGLE-ROW: column %d has dimensions %ld by %ld, current row height=%lu\n", 
	column_number,
	subgroup->width,
	subgroup->height,height_of_contents_only);
fflush(logfile);
#endif

				group_free (subgroup);
			} 
			column_number++;
		}
		w=w->next;
	}
#if 0
fprintf (logfile,"TABLE-GEO-SINGLE-ROW: row has height %lu \n", height_of_contents_only);
fflush(logfile);
#endif

	*return_row_height = height_of_contents_only;
	*return_row_width = width_of_contents_only;

	return column_number;
}


/*--------------------------------------------------------------------
 * Function to ONLY determine the number of rows and columns,
 * and then allocate needed arrays.
 *------------------------------------------------------------------*/
void
table_count_rows_and_columns (DisplayDevice *dpy, Group *current_group, 
	Word *w, Table *tbl) 
{
	int rows=0;
	int min_columns=1<<25;
	int max_columns=0;
	int i, j;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);
	CHECK_PARAM_NOT_NULL(tbl);

	while(w) {
		int tag_number = w->tag_number;

		if (tag_number == TAG_NONE) { w=w->next; continue; }

		if (tag_number == TAG_TBODY) {
			w=w->child; continue; 
		}

		if (tag_number == TAG_TR) {
			Word *w2=w->child;
			int col_count=0;

			rows++;

			while (w2) {
				if (w2->tag_number == TAG_TD) {
					char *str = word_fetch_tag_var_value (w2,"colspan");
					int value;

					if (str) {
						value = atoi(str);
						if (value >= 1) {
							col_count += value;
						}
					}
					else
						col_count++;
				}

				w2=w2->next;
			}

			if (min_columns > col_count) 
				min_columns = col_count;
			if (max_columns < col_count) 
				max_columns = col_count;
		}
		w=w->next;
	}

	tbl->rows = rows;
	tbl->min_columns = min_columns;	
	tbl->columns = max_columns;

	/* Allocate arrays to hold row/column dimensions 
	 */
	j = tbl->columns * tbl->rows;
	i = sizeof(CellInfo) * j;
	tbl->cells = (CellInfo*) malloc (i);
	if (!tbl->cells)
		error_handler ("cannot allocate table metric information");

	/* Initialize the cells */
	memset ((char*)tbl->cells, 0, i);
	for (i=0; i<j; i++) {
		tbl->cells[i].is_leading_cell=TRUE;
		tbl->cells[i].col_span=1;
		tbl->cells[i].row_span=1;
	}

	if (debug_mode && logfile)
	fprintf (logfile, "TABLE-COUNT-ROWS-COLS: total rows=%d, maxcols=%d, mincols=%d\n",
		tbl->rows, tbl->columns, tbl->min_columns);
}

/*--------------------------------------------------------------------
 * Determine a table's columns' optimal and minimal widths and
 * rows' heights.
 *------------------------------------------------------------------*/

void
table_geometry_rows_columns (DisplayDevice *dpy, Group *current_group,
  Word *w, Table *tbl)
{
	int rows, min_columns, max_columns;
	Dimension table_width, table_height, table_min_width;
	Dimension max_row_width=0;
	Dimension max_width;
	int i;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(current_group);
	CHECK_PARAM_NOT_NULL(w);
	CHECK_PARAM_NOT_NULL(tbl);

	max_width = current_group->max_width;

#if 0
fprintf(logfile,"maxwidth=%ld\n",max_width);
#endif
	rows = 0;
	min_columns = tbl->min_columns;
	max_columns = tbl->columns;

	while(w) {
		if (w->tag_number == TAG_NONE) {
			w=w->next;
			continue;
		}

		if (w->tag_number == TAG_TBODY) {
			w=w->child;
			continue;
		}

		if (w->tag_number == TAG_TR) {
			Word *w2;
			Dimension row_contents_width;
			Dimension row_contents_height;
			int total_columns;

			w2=w->child;

			/* Get column information for this row only.
			 */
			total_columns =
			  table_geometry_single_row (dpy, current_group,
			    tbl, w2, rows, 
			    &row_contents_width,
			    &row_contents_height);

			rows++;
		}
		w=w->next;
	}
}



/* Routine to process table geometry, create sub-Groups and so on.
 *
 * Table geometry is not particularly easy to process.
 * Some problems:
 * 1. A table may defined in a way that lacks column widths,
 *    either in pixels or in percentage of screen width.
 * 2. A table may be wider than the screen width, which ideally
 *    would result in it being narrowed.
 * 3. A table cannot always be narrowed to fit within the screen
 *    width, because its contents may be large but not 
 *    compressible e.g. images.
 * 4. A table may include width multiple specifications for
 *    each column, including: percentages, width as number of
 *    pixels, and the minimum width of overly wide contents.
 *
 * Processing occurs in several phases:
 * 1. The maximum and minimum number of columns and the number of
 *    rows is determined.
 * 2. The various dimensions of each cell in the table are
 *    calculated.
 * 3. Cell dimensions are analyzed to determine the best
 *    table size.
 * 4. Each cell is realized as a new Group and contents are
 *    given coordinates.
 */

int
table_geometry (DisplayDevice *dpy, Group *current_group, Word *w, Table *tbl) 
{
	Word *w2;
	int rows, min_columns, max_columns;
	int i,j;
	Dimension max_width;
	Dimension table_width; /* table_height; */
	Coordinate x, y;
	Dimension *rough_widths;
	unsigned char *rough_widths_remains_fixed;
	Dimension *rough_heights;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);
	CHECK_PARAM_NOT_NULL(tbl);

	/* Determine our horizontal limit */
	/* XX no account taken of cell spacing & padding */

	/**********************************************************
	 * STEP 0: Get the number of rows and columns
	 *********************************************************/

	table_count_rows_and_columns (dpy, current_group, w, tbl);

	rows = tbl->rows;
	min_columns = tbl->min_columns;
	max_columns = tbl->columns;

	if (!rows)
		error_handler ("table has no rows");
	if (!max_columns)
		error_handler ("table has no columns");

	/**********************************************************
	 * STEP 1: Calculate maximum width of all columns' contents.
         * This is just an optimistic number really.
	 *********************************************************/

	tbl->max_content_width = current_group->max_width - 2 * tbl->border_size;
	tbl->max_content_width -= tbl->cell_spacing * (tbl->columns-1);
	tbl->max_content_width -= tbl->cell_padding * 2 * tbl->columns;
	max_width = current_group->max_width;

	/**********************************************************
	 * STEP x: Adjust desired width if necessary.
	 *
	 * XX - not appropriate, since the HTML could specifically
	 * ask for a table wider than the screen and we should
	 * honor that request.
	 *********************************************************/
	
#if 0
	if (tbl->desired_width && tbl->desired_width > max_width)
		tbl->desired_width = max_width;
#endif

	/**********************************************************
	 * STEP 2: Allocate arrays and determine best-case, minimum
	 * and desired column and row sizes.
	 *********************************************************/

	table_geometry_rows_columns (dpy,current_group,w,tbl);

	/**********************************************************
	 * STEP 3: Analyze the optimal, minimal, and desired column
	 * widths to come up with a rough set of column widths, taking
	 * the maximum width into account as well.
	 *********************************************************/

	rough_widths = (Dimension*) my_malloc (sizeof (Dimension) * max_columns);
	if (!rough_widths)
		error_handler ("unable to allocate space for processing of table");

	tbl->column_widths = rough_widths;

	rough_widths_remains_fixed = (unsigned char*) my_malloc (max_columns);
	if (!rough_widths_remains_fixed)
		error_handler ("unable to allocate space for processing of table");

	for (i=0; i<max_columns; i++) {
		Dimension max_min_wid = 0; /* maximum of all minimal_width values */
		Dimension max_desired = 0; /* maximum of all desired_width values */
		Dimension max_optimal = 0; /* maximum of all optimal_width values */
		Dimension rough;

		for (j=0; j<rows; j++) {
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			int span;

			span = cell->col_span;

			if (cell->is_leading_cell && span==1) 
			{
				if (cell->desired_width > max_desired)
					max_desired = cell->desired_width;
				if (cell->minimal_width > max_min_wid)
					max_min_wid = cell->minimal_width;
				if (cell->optimal_width > max_optimal)
					max_optimal = cell->optimal_width;
			}
		}

		if (max_min_wid > max_width) {
/* need to return False to tell main() to increase max_width */
/* XX - there should be a better way */
			error_handler ("table processing: maximum minimal_width is wider than maximum width of Group");
		}
		if (max_optimal > max_width)
			max_optimal = max_width;
		if (max_desired > max_width)
			max_desired = max_width;

		rough = max_min_wid;
		if (max_desired) {
#if 0
			if (max_optimal < max_desired) {
				fprintf (logfile, "max_optimal < max_desired -- good\n");
			} else {
				fprintf (logfile, "max_optimal > max_desired -- bad, but ok\n");
			}
#endif
			if (max_desired > rough) 
				rough = max_desired;
		}
		if (max_optimal > rough)
			rough = max_optimal;

		rough_widths[i] = rough;
		rough_widths_remains_fixed[i] = max_desired ? TRUE : FALSE; 
	}

	/**********************************************************
	 * STEP 4: Now pare things down to fit on screen.
	 *********************************************************/
	{
		Dimension total_width;
		Dimension total_width_min;
		int total_variable=max_columns;

		total_width = 0;
		total_width_min = 0;

		for (i=0; i<max_columns; i++) {
			if (rough_widths_remains_fixed[i]) {
				total_width_min += rough_widths[i];
				total_variable--;
			}

			total_width += rough_widths[i];
		}

#if 0
fprintf(logfile,"total width=%ld\n", total_width);
fprintf(logfile,"total variable=%ld\n", total_variable);
#endif

		if (total_width_min > tbl->max_content_width) {
			error_handler ("table contents are too wide to fit on screen");
		}

		/* If the columns are asking for too much, 
		 * we need to limit them.
		 */
		if (total_width > tbl->max_content_width) {
			Dimension variable_amount;
			Dimension chunk;

			if (!total_variable)
				error_handler ("table contents are too wide to fit on screen");

			variable_amount = tbl->max_content_width - total_width_min;
			chunk = variable_amount / total_variable;

			for (i=0; i<max_columns; i++) {
				if (!rough_widths_remains_fixed[i]) {
					rough_widths[i] = chunk;
					variable_amount -= chunk;
					total_variable--;
					if (total_variable==0)
						rough_widths[i] += variable_amount;
				}
			}
		}
	}

#if 0
	/**********************************************************
	 * STEP 3B: Store the rough widths into the cells.
	 *********************************************************/

	for (i=0; i<max_columns; i++) {
		Dimension rough;

		for (j=0; j<rows; j++) {
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			if (cell->is_leading_cell) 
			{
			}
		}
	}
#endif

	/**********************************************************
	 * STEP 4: Now let us process the columns which include
	 * a column span request.
	 *********************************************************/

#if 0
	for (i=0; i<max_columns; i++) {
		fprintf(logfile, "Column width before colspan processing: #%d = %ld\n",
			i, rough_widths[i]);
	}
#endif

	for (j=0; j<rows; j++) {
		for (i=0; i<max_columns; i++) {
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			int span;

			span = cell->col_span;

			if (cell->is_leading_cell && span > 1) 
			{
				Dimension cumulative_width = 0;
				int k;

				for (k=0; k<span; k++) {
					cumulative_width += rough_widths[k+i];
				}

				/* If the cumulative width is less than
				 * the width of our multiple-column cell,
				 * then we need to increase the rough_width
				 * values evenly.
				 */
				if (cumulative_width < cell->optimal_width) {
					Dimension diff = cell->optimal_width -
						cumulative_width;
					Dimension chunk;

					chunk = diff / span;

					for (k=0; k<span; k++) {
						rough_widths[k+i] += chunk;
						diff -= chunk;
					}
					if (diff)
						rough_widths[i] += diff;
				}
			}
		}
	}

#if 0
	for (i=0; i<max_columns; i++) {
		fprintf(logfile, "Column width after colspan processing: #%d = %ld\n",
			i, rough_widths[i]);
	}
#endif

	/**********************************************************
	 * STEP 5: Now that we have reliable column widths, we can
	 * create the children Groups. This is necessary before we
	 * can deal with row heights, since heights may change once
	 * the new widths are used.
	 *********************************************************/

#if 0
	y = tbl->border_size;
	y += tbl->cell_padding;
#endif
	y=0;

	for (j=0; j<rows; j++) 
	{
		Dimension rough_height=0;

		x = tbl->border_size;
		x += tbl->cell_padding;

		for (i=0; i<max_columns; i++) 
		{
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			Group *subgroup;
			int span;

			span = cell->col_span;

			if (cell->is_leading_cell)
			{
				Dimension cumulative_width = 0;
				int k;
				Word *kids;

				/* Determine the width of this cell, given
				 * its span which can be 1 or larger.
				 */
				for (k=0; k<span; k++) {
					cumulative_width += rough_widths[k+i];
				}

				/* We're all set to create the child.
				 * Note, process_table_tag created a parent
				 * Group for the entire table, and that's
				 * our current_group.
				 */
				subgroup=group_new (dpy, current_group, 
					0,0,
					cumulative_width,MAX_COORD);

				kids = cell->td->child;
				if (kids)
				  group_geometry (dpy, subgroup, kids);

				/* Very importantly, here we store the
				 * pointer to the sub-Group for this cell.
				 * It's needed below for setting the y
				 * value, but also for later writing out
				 * of HTML.
				 */
				cell->td->subgroup = (void*) subgroup;

#if 0
				if (subgroup->height > rough_height)
					rough_height = subgroup->height;
#endif
			}

			x += rough_widths [i];
			x += tbl->cell_padding; /* rt side */
			x += tbl->cell_spacing; /* between cells */
			x += tbl->cell_padding; /* lt side */
		}

#if 0
		y += rough_height;
		y += tbl->cell_padding; /* rt side */
		y += tbl->cell_spacing; /* between cells */
		y += tbl->cell_padding; /* lt side */
#endif
	}

	/**********************************************************
	 * STEP 6: Now let us analyze the cell heights to get rough
	 * row height values, needed before we can consider any
	 * rowspan requests.
	 *********************************************************/

	rough_heights = (Dimension*) my_malloc (sizeof (Dimension) * rows);
	if (!rough_heights)
		error_handler ("unable to allocate space for processing of table");

	tbl->row_heights = rough_heights;

	for (j=0; j<rows; j++) 
	{
		Dimension max_desired = 0; /* maximum of all desired_height values */
		Dimension max_optimal = 0; /* maximum of all optimal_height values */
		Dimension rough;

		for (i=0; i<max_columns; i++) 
		{
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			int span;

			span = cell->row_span;

			if (cell->is_leading_cell && span==1) 
			{
				if (cell->desired_height > max_desired)
					max_desired = cell->desired_height;
				if (cell->optimal_height > max_optimal)
					max_optimal = cell->optimal_height;
			}
		}

#if 0
		if (max_min_wid > max_height) {
/* need to return False to tell main() to increase max_width */
/* XX - there should be a better way */
			error_handler ("table processing: maximum minimal_width is wider than maximum width of Group");
		}
		if (max_optimal > max_height)
			max_optimal = max_height;
		if (max_desired > max_height)
			max_desired = max_height;
#endif

		rough = max_optimal;
		if (max_desired) {
#if 0
			if (max_optimal < max_desired) {
				fprintf (logfile, "height: max_optimal < max_desired -- good\n");
			} else {
				fprintf (logfile, "height: max_optimal > max_desired -- bad, but ok\n");
			}
#endif
			if (max_desired > rough) 
				rough = max_desired;
		}

		rough_heights[j] = rough;
	}

#if 0
i=0;
for (j=0; j<rows; j++) {
fprintf (logfile, "rough_heights [%d]=%ld\n", j, rough_heights [j]);
i += rough_heights [j];
}
i += tbl->border_size*2;
i += tbl->cell_padding*rows*2;
i += tbl->cell_spacing*(rows-1);
#endif

	/**********************************************************
	 * STEP 7: Now let us process the rows which include
	 * a row span request.
	 *********************************************************/

	for (j=0; j<rows; j++) 
	{
		for (i=0; i<max_columns; i++) 
		{
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			int span;

			span = cell->row_span;

			if (cell->is_leading_cell && span > 1) 
			{
				Dimension cumulative_height = 0;
				int k;

				for (k=0; k<span; k++) {
					cumulative_height += rough_heights[k+j];
				}

				/* If the cumulative height is less than
				 * the height of our multiple-row cell,
				 * then we need to increase the rough_height
				 * values evenly.
				 */
				if (cumulative_height < cell->optimal_height) 
				{
					Dimension diff = cell->optimal_width -
						cumulative_height;
					Dimension chunk;

					chunk = diff / span;

					for (k=0; k<span; k++) {
						rough_heights[k+i] += chunk;
						diff -= chunk;
					}
					if (diff)
						rough_heights[i] += diff;
				}
			}
		}
	}

#if 0
	for (i=0; i<rows; i++) {
		fprintf(logfile, "Row height after rowspan processing: #%d = %ld\n",
			i, rough_heights [i]);
	}
#endif

	/**********************************************************
	 * STEP 8: Now that we have BOTH reliable column widths and
	 * reliable row heights, we can assign x and y values to child
	 * Groups.
	 *********************************************************/

	y = tbl->border_size;
	y += tbl->cell_padding;

	for (j=0; j<rows; j++) 
	{
		x = tbl->border_size;
		x += tbl->cell_padding;

		for (i=0; i<max_columns; i++) 
		{
			CellInfo *cell = &tbl->cells [max_columns * j + i];
			Group *subgroup;
			int span;

			if (cell->is_leading_cell)
			{
				Word *child;
				Dimension cumulative_width = 0;
				Dimension cumulative_height = 0;
				int k;

				for (k=0; k < cell->col_span; k++) {
					cumulative_width += rough_widths[k+i];
				}

				for (k=0; k < cell->row_span; k++) {
					cumulative_height += rough_heights[k+j];
				}

				/* We're all set to position the child.
				 */
				subgroup= (Group*) cell->td->subgroup;

				subgroup->x0 += x;
				subgroup->y0 += y;
				// subgroup->max_width = cumulative_width;
				// subgroup->max_height =cumulative_height;

				child = cell->td->child;
#if 0
				if (child)
				  word_apply_offset (child, x, y);
#endif
#if 0
fprintf (logfile, "post-apply: cell->td->x=%ld\n", cell->td->x);
fprintf (logfile, "cell->td->str=%s\n", cell->td->str);
fprintf (logfile, "x=%ld\n", x);
fprintf (logfile, "spacing=%ld\n", tbl->cell_spacing);
fprintf (logfile, "padding=%ld\n", tbl->cell_padding);
#endif
			}

			x += rough_widths [i];
			x += tbl->cell_padding; /* rt side */
			x += tbl->cell_spacing; /* between cells */
			x += tbl->cell_padding; /* lt side */
		}

		y += rough_heights [j];
		y += tbl->cell_padding; /* bottom side */
		y += tbl->cell_spacing; /* between cells */
		y += tbl->cell_padding; /* top side */
	}

	/* Remove extra */
	x -= tbl->cell_spacing; 
	x -= tbl->cell_padding; 
	y -= tbl->cell_spacing; 
	y -= tbl->cell_padding; 

	/* Add on the right/bottom borders */
	x += tbl->border_size; 
	y += tbl->border_size; 

	tbl->overall.optimal_width = x;
	tbl->overall.minimal_width = x;
	tbl->overall.optimal_height = y;

	/**********************************************************
	 * STEP 9: Update the table-wide values.
	 *********************************************************/

	tbl->x0 = 0;
	tbl->y0 = 0;
	tbl->x1 = tbl->overall.optimal_width - 1;
	tbl->y1 = tbl->overall.optimal_height - 1;

	/**********************************************************
	 * STEP 10: Clean up.
	 *********************************************************/

	my_free (rough_widths_remains_fixed);

	return TRUE;
}


/*========================Tag Execution Routines=========================*/

/*-------Paragraphs & Lines----------------------------------------------*/

int
process_para_tag (DisplayDevice *dpy, Group *current_group, Word *w, 
  int list_type) 
{
	int retval=TRUE;
	/* unsigned char align=ALIGN_LEFT; */
	Group *subgroup;
	Dimension height;
	Dimension padding;
	Dimension newwidth;
	String str;

	group_end_line (dpy, current_group);

	if (!current_group->beginning_new_paragraph) {
		padding = 7;
	} else {
		padding = 0;
	}

	subgroup = group_new (dpy, current_group, 
		0,
		current_group->ystart+ padding ,  /* pre-paragraph spacing */
		current_group->max_width - (current_group->x0+current_group->x),
		current_group->max_height);

	w->subgroup = (void*) subgroup;

	str = word_fetch_tag_var_value (w,"bgcolor");
	if (str) {
		subgroup->bgcolor = color_lookup (str);
	}

	subgroup->beginning_new_paragraph = TRUE;

	if (w->child) {
		if (!group_geometry (dpy, subgroup, w->child))
			retval=FALSE;
	}

	height = subgroup->height;
	height += padding;

	w->width = w->min_width = subgroup->width;
	w->ascent = height;
	w->descent = 0;
	group_append_word (dpy, current_group, w);

	current_group->beginning_new_paragraph = FALSE;
	current_group->insert_paragraph_space = TRUE;
#if 0
	current_group->ystart += height;
	current_group->height += height;

	newwidth = subgroup->width + current_group->xstart;

	if (newwidth > current_group->width)
		current_group->width = newwidth;
#endif

	return retval;
}


int
process_a_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	char *str;
	int retval=TRUE;

	ENTER;

	group_append_word (dpy, current_group, w);

	/* Only print this tag's children if it has an HREF param.
	 * Tags which are NAME= only are not to be printed.
	 */
	str = word_fetch_tag_var_value (w,"href");
	if (str) {
		Style *sl = style_new();
		sl->flags = ATTR_LINK;
		sl->name = my_strdup ("a");

		group_push_style (dpy, current_group, sl);

		if (w->child) {
			if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			  
				retval = FALSE;
		}

		group_drop_style (current_group);
	}

	return retval;
}


int
process_ignore_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval = TRUE;

	ENTER;

	/* Whatever <script> contains is ignored. */

	return retval;
}

int
process_passthrough_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval = TRUE;

	ENTER;

	/* Used with "noscript" and "q" tags.
	 */

	if (w->child) {
		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval = FALSE;
	}

	return retval;
}


int
process_break_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	ENTER;

	/* Put the actual BR tag into the line, since there could be
	 * any number of them. This we cannot handle with flags as with P.
	 */
	w->width=0;
	w->ascent= group_get_current_ascent (dpy, current_group);
	w->descent=0;
	group_append_word (dpy, current_group, w);
	group_end_line_make_line(dpy, current_group);

	return retval;
}


int
process_hr_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) 
{
	char *str;
	int desired_height;
	int total_height;

	ENTER;

	str = word_fetch_tag_var_value (w,"size");
	if (str) {
		desired_height = atoi(str);
		if (!desired_height) 
			desired_height=1;
		else if (desired_height > 100) 
			desired_height = 100;
	} else
		desired_height=1;

	total_height = desired_height + dpy->horizontal_line_space;

	group_end_line (dpy, current_group);
	group_make_spacer_line (current_group, total_height);

	w->width = 1; /* XX */
	w->ascent = total_height;
	w->descent = 0;
	w->tag.datum = desired_height;

	group_append_word (dpy, current_group, w);

	group_end_line_make_line (dpy, current_group);

	return TRUE;
}


/*-------Images----------------------------------------------------------*/
int
process_img_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) 
{
	char *str;
	char *str2;
	char *str3;
	Dimension wid,ht;

	ENTER;

	str = word_fetch_tag_var_value (w,"src");
	str2 = word_fetch_tag_var_value (w,"width");
	str3 = word_fetch_tag_var_value (w,"height");

	if (str2) wid = atoi (str2); else wid=INVALID_DIMENSION;
	if (str3) ht = atoi (str3); else ht=INVALID_DIMENSION;

	if (debug_mode && logfile) {
		fprintf (logfile, "IMAGE, wid=%s, ht=%s, src=%s\n", 
			wid==INVALID_DIMENSION ? "(none)" : str2,
			ht==INVALID_DIMENSION ? "(none)" : str3,
			str ? str : "(none)");
		fflush(logfile);
	}

	if (dpy->width < 200) {
		w->width=3;
		w->min_width=3;
		w->ascent=3;
		w->descent=0;
	}
	else
	{
		if (!strncmp (str, "http://", 7)) 
		{
		} 
		else 
		{
			String path;
			String extn;
			String tmp;

			if (!strncmp (str, "file:", 5)) 
				path = &str[5];
			else
				path = str;

			/* Convert DOS paths of the A|file.htm variety.
			 */
			tmp = strchr (path, '|');
			if (tmp) *tmp=':';

			/* Determine if the file is an XBM file.
			 */
			extn = path;
			do {
				tmp = strchr (extn,'.');
				if (tmp) extn = tmp + 1;
			} while (tmp);

			if (debug_mode && logfile) {
				fprintf(logfile,"IMG: file extension = %s\n",extn);
				fflush(logfile);
			}

			if (!strcmp ("xbm", extn) ||
			    !strcmp ("XBM", extn))
			{
				FILE *f;
				Image *im = image_new();

				f=fopen(path,"r");
				if (!f) {
					warning_handler ("cannot open local XBM image file");
				} else {
					if (!image_read_xbm (im, f, dpy->fg, dpy->bg))
						warning_handler ("local XBM image file contains invalid data");
					else {
						w->tag.ptr = (void*) im;
						wid=im->width;
						ht=im->height;
						if (debug_mode && logfile) {
							fprintf(logfile, "XBM image read was successful\n");
							fflush(logfile);
						}
					}
				}
			}
		}

		if (wid == INVALID_DIMENSION) wid=11;
		if (ht == INVALID_DIMENSION) ht=11;

		w->width=wid;
		w->min_width=wid;
		w->ascent=ht;
		w->descent=0;
	}

	group_append_word (dpy, current_group, w);

	return TRUE;
}



/*-------Drawn Images------------------------------------------------------*/


static int 
parse_six_coords (String str, 
	Coordinate *return_x0,
	Coordinate *return_y0,
	Coordinate *return_z0,
	Coordinate *return_x1,
	Coordinate *return_y1, 
	Coordinate *return_z1) 
{
	CHECK_PARAM_NOT_NULL(str);
	CHECK_PARAM_NOT_NULL(return_x0);
	CHECK_PARAM_NOT_NULL(return_x1);
	CHECK_PARAM_NOT_NULL(return_y0);
	CHECK_PARAM_NOT_NULL(return_y1);
	CHECK_PARAM_NOT_NULL(return_z0);
	CHECK_PARAM_NOT_NULL(return_z1);

	if (*str=='x' || *str=='X') {
		str++;
		if (strlen(str) >= 8) {
			*return_x0 = c2toi_signed (str); str+=2;
			*return_y0 = c2toi_signed (str); str+=2;
			*return_z0 = c2toi_signed (str); str+=2;
			*return_x1 = c2toi_signed (str); str+=2;
			*return_y1 = c2toi_signed (str); str+=2;
			*return_z1 = c2toi_signed (str);
			return TRUE;
		}
	} else {
		if (6 == sscanf (str, "%ld,%ld,%ld,%ld,%ld,%ld", 
		                 return_x0, return_y0, return_z0,
		                 return_x1, return_y1, return_z1))
			return TRUE;
	}
	return FALSE;
}

static int 
parse_four_coords (String str, 
	Coordinate *return_x0,
	Coordinate *return_y0,
	Coordinate *return_x1,
	Coordinate *return_y1) 
{
	CHECK_PARAM_NOT_NULL(str);
	CHECK_PARAM_NOT_NULL(return_x0);
	CHECK_PARAM_NOT_NULL(return_x1);
	CHECK_PARAM_NOT_NULL(return_y0);
	CHECK_PARAM_NOT_NULL(return_y1);

	if (*str=='x' || *str=='X') {
		str++;
		if (strlen(str) >= 8) {
			*return_x0 = c2toi_signed (str);
			*return_y0 = c2toi_signed (str+2);
			*return_x1 = c2toi_signed (str+4);
			*return_y1 = c2toi_signed (str+6);
			return TRUE;
		}
	} else {
		if (4 == sscanf (str, "%ld,%ld,%ld,%ld", return_x0, return_y0,
							 return_x1, return_y1))
			return TRUE;
	}
	return FALSE;
}


static int 
parse_three_coords (String str, 
	Coordinate *return_a,
	Coordinate *return_b,
	Coordinate *return_c)
{
	CHECK_PARAM_NOT_NULL(str);
	CHECK_PARAM_NOT_NULL(return_a);
	CHECK_PARAM_NOT_NULL(return_b);
	CHECK_PARAM_NOT_NULL(return_c);

	if (*str=='x' || *str=='X') {
		str++;
		if (strlen(str) >= 6) {
			*return_a = c2toi_signed (str);
			*return_b = c2toi_signed (str+2);
			*return_c = c2toi_signed (str+4);
			return TRUE;
		}
	} else {
		if (3 == sscanf (str, "%ld,%ld,%ld", return_a, return_b, return_c))
			return TRUE;
	}
	return FALSE;
}



static int 
parse_two_coords (String str, 
	Coordinate *return_a,
	Coordinate *return_b)
{
	CHECK_PARAM_NOT_NULL(str);
	CHECK_PARAM_NOT_NULL(return_a);
	CHECK_PARAM_NOT_NULL(return_b);

	if (*str=='x' || *str=='X') {
		str++;
		if (strlen(str) >= 4) {
			*return_a = c2toi_signed (str);
			*return_b = c2toi_signed (str+2);
			return TRUE;
		}
	} else {
		if (2 == sscanf (str, "%ld,%ld", return_a, return_b))
			return TRUE;
	}
	return FALSE;
}


/* This tag is named "draw" because it's the subtags instruct Beest to make
 * a drawing, which is a one-time operation the result of which is an Image.
 * This is to be contrasted with the "model" tag, which produces a 3-D model
 * which can be interactively rotated.
 *
 * Example use of this tag (no compression):
 *
 * <DRAW RANGE=0,99,0,99 BGCOLOR=gray2>
 *   <PIXEL COLOR=red VECTORS=0,50>
 *   <LINE COLOR=red VECTORS=0,50,50,99>
 *   <LINEREL COLOR=red VECTORS=0,50>
 *   <RECT COLOR=green VECTORS=0,0,99,99>
 *   <HLINE COLOR=black VECTORS=0,99,50>   <!-- x0,x1,y -->
 *   <VLINE COLOR=black VECTORS=50,0,99>   <!-- x,y0,y1 -->
 *   <FILLEDRECT COLOR=blue BGCOLOR=white VECTORS=25,25,75,75>
 *  Not implemented:
 *   <OVAL COLOR=green VECTORS=0,0,99,99>  <!-- x0,y0,x1,y1 -->
 *   <FILLEDOVAL COLOR=green bgcolor=black VECTORS=0,0,99,99>
 * </DRAW>
 *
 * The tags themselves all have shorter variants except DRAW
 * and the oval tags. These are as follows:
 *
 *   PIXEL = P~
 *   LINE = L~
 *   LINEREL = LR~
 *   RECT = R~
 *   HLINE = H~
 *   VLINE = V~
 *   FILLEDRECT = F~
 * 
 * The COLOR and VECTORS parameters have shorter variants too:
 *
 *   COLOR = C
 *   VECTORS = V
 * 
 * The vector information can be compressed as follows. All coordinates
 *   are limited to the range -2048 through 2047. Due to this, each fits
 *   neatly into a 12-bit value which itself fits into two compressed 
 *   ASCII bytes.
 * To compress some coordinate data, convert it to hexadecimal as in
 *   the following example:
 *  <L~ C=red V='x. . G2G2'>
 * This draws a red line from (64,64) to (300,300)
 * The uncompressed equivalent is a bit longer:
 *  <LINE COLOR=red VECTORS=64,64,300,300>
 *
 * Compression of one coordinate is as follows: (see also test/draw.c)
 *   It's a 12-bit value separated into two 6-bit values.
 *   The upper 6-bits appear first, then the lower.
 *   One six bit value converts as follows:
 *     0 = ' '
 *     1 = '.'
 *     2..11 = '0'..'9'
 *     12..37 = 'a'..'z'
 *     38..63 = 'A'..'Z'
 *
 * Some general things to know about the DRAW tag:
 *   0. non-tags (textual words) plus non-drawing related tags are ignored
 *   1. all the tag names are singular, e.g. there is no <LINES>
 *   2. the LINE tag however allows for connected segments
 *   3. vectors are two-dimensional hence there's only an x,y pair
 *   4. coordinates are in base10 notation for readability
 *   5. DRAW should not be nested
 *   6. no line width parameter yet
 *   7. ovals are not implemented
 *   8. maximum coordinates are in the range [-2048,2047] 
 *   9. in the compressed format, a coordinate is 12-bit signed hexadecimal
 *   10. the ranges value can be compressed
 *   11. the compression rate is between about 30% and 50%
 */

int
process_draw_tag (DisplayDevice *dpy, Group *current_group, Word *w, 
                int list_type) 
{
	int retval=TRUE;
	Word *w2;
	Dimension height;
	Dimension width;
	Coordinate ix0,iy0,ix1,iy1;
	String str;
	Image *im;
	Coordinate prev_x, prev_y;

	ENTER;

	if (debug_mode && logfile) {
		fprintf(logfile, "ENTERED DRAW TAG\n");
		fflush(logfile);
	}
	
	/* The RANGES parameter is mandatory.
	 */
	str = word_fetch_tag_var_value (w,"ranges");
	if (!str) {
		warning_handler ("Drawn image is missing a ranges param.");
		return TRUE;
	}
	if (!parse_four_coords (str, &ix0,&ix1,&iy0,&iy1)) {
		warning_handler ("Drawn image has incorrectly formatted ranges param.");
		return TRUE;
	}

	if (ix0>ix1 || iy0>iy1 || ix0>2047 || ix1<-2048 || iy0>2047 || 
	    iy1<-2048) {
		warning_handler ("Drawn image has ranges values -- tag ignored.");
		return TRUE;
	}
	if (ix0<-2048) ix0=-2048;
	if (ix1>2047) ix1=2047;
	if (iy0<-2048) iy0=-2048;
	if (iy1>2047) iy1=2047;

	width=ix1-ix0+1;
	height=iy1-iy0+1;

	/* XX - Notice: no border is being used */

	w->width = width;
	w->min_width = width;
	w->ascent = height;
	w->descent = 0;

	im = image_new ();
	image_init (im, width, height);

	{
		String str = word_fetch_tag_var_value (w,"bgcolor");
		int pixel;
		
		if (str) {
			Color c = color_lookup (str);

			pixel = dpy->rgb_to_pixel (c);

			if (pixel < 0) {
				/* XX -- allocate, or get greyscale */
				pixel = dpy->bg;
			}
		}
		else
			pixel = dpy->bg;

		image_clear (im, pixel);
	}

	w->tag.ptr = (void*) im;

	prev_x = prev_y = INVALID_COORDINATE;

	w2 = w->child;
	while (w2) 
	{
		if ('<'==*w2->str && w2->tag_number != TAG_NONE) 
		{
			Coordinate x0,y0,x1,y1;
			int pixel = -1;
			Coordinate color;
			String vectors;
			String colorstr;

			vectors= word_fetch_tag_var_value (w2,"v");
			if (!vectors)
				vectors = word_fetch_tag_var_value (w2,"vectors");

			colorstr = word_fetch_tag_var_value (w2,"c");
			if (!colorstr)
				colorstr = word_fetch_tag_var_value (w2,"color");

			if (colorstr) {
				color=color_lookup (colorstr);
				if (color != INVALID_COLOR) {
					pixel = dpy->rgb_to_pixel (color);
				}
				if (debug_mode && logfile) {
					fprintf (logfile, "DRAW_TAG: tag %s will use color %s, which is pixel %d\n",
						w->segments->str, colorstr, pixel);
					fflush(logfile);
				}
			}
			if (pixel < 0) 
				pixel = dpy->fg;

			switch (w2->tag_number) {
			case TAG_PIXEL: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-pixel vector=%s\n", vectors);
					fflush(logfile);
				}

				if (vectors && parse_two_coords (vectors,&x0,&y0)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					image_draw_pixel (im,x0,y0,pixel);
					prev_x = x0;
					prev_y = y0;
				}
				break;
			case TAG_LINEREL: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-line-relative vector=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_two_coords (vectors,&x0,&y0)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					if (prev_x != INVALID_COORDINATE &&
					    prev_y != INVALID_COORDINATE)
						image_draw_line (im,prev_x,prev_y,x0,y0,pixel);
					prev_x = x0;
					prev_y = y0;
				}
				break;
			case TAG_LINE: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-line vectors=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_four_coords (vectors,&x0,&y0,&x1,&y1)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					x1 -= ix0;
					y1 -= iy0;
					image_draw_line (im,x0,y0,x1,y1,pixel);
					prev_x = x1;
					prev_y = y1;
				}
				break;
			case TAG_HLINE: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-horizontal-line vectors=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_three_coords (vectors,&x0,&x1,&y0)) 
				{
					x0 -= ix0;
					x1 -= ix0;
					y0 -= iy0;
					image_draw_hline (im,x0,x1,y0,pixel);
					prev_x = x1;
					prev_y = y0;
				}
				break;
			case TAG_VLINE: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-vertical-line vectors=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_three_coords (vectors,&x0,&y0,&y1)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					y1 -= iy0;
					image_draw_vline (im,x0,y0,y1,pixel);
					prev_x = x0;
					prev_y = y1;
				}
				break;
			case TAG_RECT: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-rect vectors=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_four_coords (vectors,&x0,&y0,&x1,&y1)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					x1 -= ix0;
					y1 -= iy0;
					image_draw_rect (im,x0,y0,x1,y1,pixel);
					prev_x = x1;
					prev_y = y1;
				}
				break;
			case TAG_FILLEDRECT: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-filled-rect vectors=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_four_coords (vectors,&x0,&y0,&x1,&y1)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					x1 -= ix0;
					y1 -= iy0;
					image_draw_filled_rect (im,x0,y0,x1,y1,pixel);
					prev_x = x1;
					prev_y = y1;
				}
				break;

			case TAG_TEXT: 	{
				Style sl;
				String expr, face, sizestr, italic, bold;
				FontSize size;

				style_reset(&sl);

				if (debug_mode && logfile) {
					fprintf (logfile, "draw-text vectors=%s\n", vectors);
					fflush(logfile);
				}

				expr= word_fetch_tag_var_value (w2,"expr");
				face= word_fetch_tag_var_value (w2,"face");
				sizestr= word_fetch_tag_var_value (w2,"size");
				if (sizestr) 
					size=atoi(sizestr);
				else 
					size=INVALID_FONTSIZE;
				italic= word_fetch_tag_var_value (w2,"italic");
				bold= word_fetch_tag_var_value (w2,"bold");

				if (face) sl.family=my_strdup(face);
				sl.size=size;
				sl.flags= (italic?ATTR_ITALIC:0) |
				          (bold?ATTR_BOLD:0);

				/* XX - Need to process inline Style here
				 */

				if (debug_mode && logfile) {
					fprintf(logfile,"DRAW: text ");
					fflush(logfile);
					style_dump(logfile,&sl);
					fprintf(logfile,"\n");
					fflush(logfile);
				}

				style_fix_font (dpy, &sl);

				if (vectors && parse_two_coords (vectors,&x0,&y0)) 
				{
					x0 -= ix0;
					y0 -= iy0;
					image_draw_text (im,
						sl.fontdata,
						sl.flags & ATTR_BOLD,
						sl.flags & ATTR_ITALIC,
						expr,x0,y0,pixel);
					prev_x = x0;
					prev_y = y0;
				}

				if (face) my_free(face);
				break;
			}

			case TAG_OVAL: 
				if (debug_mode && logfile) {
					fprintf (logfile, "NO-OP: draw-oval vectors=%s\n", vectors);
					fflush(logfile);
				}
				break;
			case TAG_FILLEDOVAL: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "NO-OP: draw-filled-oval vectors=%s\n", vectors);
					fflush(logfile);
				}
				break;
			} /* switch */
		}
		w2=w2->next;
	}

	group_append_word (dpy, current_group, w);

	return retval;
}



/* 3-D support
 *
 * The MODEL tag results in an interactive image of a 3-D image.
 * It uses several of the subtags that DRAW does except the filled
 * forms.
 */

int
process_model_tag (DisplayDevice *dpy, Group *current_group, Word *w, 
                int list_type) 
{
	int retval=TRUE;
	Word *w2;
	Dimension width, max_dimension, min_dimension;
	Coordinate ix0,iy0,iz0,ix1,iy1,iz1;
	String str;
	Image *im;
	Coordinate prev_x, prev_y, prev_z;
	Coordinate angle_x, angle_y, angle_z;

	ENTER;

	if (debug_mode && logfile) {
		fprintf(logfile, "ENTERED MODEL TAG\n");
		fflush(logfile);
	}
	
	/* The RANGES parameter is mandatory.
	 */
	str = word_fetch_tag_var_value (w,"ranges");
	if (!str) {
		warning_handler ("Model is missing a ranges param.");
		return TRUE;
	}
	if (!parse_six_coords (str, &ix0,&ix1,&iy0,&iy1,&iz0,&iz1)) {
		warning_handler ("Model has incorrectly formatted ranges param.");
		return TRUE;
	}

	str = word_fetch_tag_var_value (w,"angles");
	if (!str) {
		angle_x = angle_y = angle_z = 0;
	} else {
		if (!parse_three_coords (str,&angle_x,&angle_y,&angle_z))
		{
			warning_handler ("Model has incorrectly formatted angles param.");
			return TRUE;
		}
	}

	if (ix0>ix1 || iy0>iy1 || iz0>iz1 || ix0>2047 || ix1<-2048 || iy0>2047 || 
	    iy1<-2048 || iz0>2047 || iz1<-2048) {
		warning_handler ("Drawn image has ranges values -- tag ignored.");
		return TRUE;
	}

	if (ix0<-2048) ix0=-2048;
	if (ix1>2047) ix1=2047;
	if (iy0<-2048) iy0=-2048;
	if (iy1>2047) iy1=2047;
	if (iz0<-2048) iz0=-2048;
	if (iz1>2047) iz1=2047;

	if (angle_x < -360) angle_x = -360;
	if (angle_x > 360) angle_x = 360;
	if (angle_y < -360) angle_y = -360;
	if (angle_y > 360) angle_y = 360;
	if (angle_z < -360) angle_z = -360;
	if (angle_z > 360) angle_z = 360;
	if (angle_x < 0) angle_x += 360;
	if (angle_y < 0) angle_y += 360;
	if (angle_z < 0) angle_z += 360;

	min_dimension = ix0;
	if (iy0 < min_dimension) min_dimension = iy0;
	if (iz0 < min_dimension) min_dimension = iz0;

	max_dimension = ix1;
	if (iy1 > max_dimension) max_dimension = iy1;
	if (iz1 > max_dimension) max_dimension = iz1;

	/* If the user has put the object on one side of an
	 * axis (bad!) we can still work with it, but the
 	 * Image with be much larger.
	 */
	if (min_dimension > 0) min_dimension=-max_dimension;
	if (max_dimension < 0) max_dimension=-min_dimension;

	width=max_dimension - min_dimension;

	/* XX - Notice: no border is being used */

	w->width = width;
	w->min_width = width;
	w->ascent = width;
	w->descent = 0;

	im = image_new ();
	image_init (im, width, width);

	{
		String str = word_fetch_tag_var_value (w,"bgcolor");
		int pixel;
		
		if (str) {
			Color c = color_lookup (str);

			pixel = dpy->rgb_to_pixel (c);

			if (pixel < 0) {
				/* XX -- allocate, or get greyscale */
				pixel = dpy->bg;
			}
		}
		else
			pixel = dpy->bg;

		image_clear (im, pixel);
	}

	w->tag.ptr = (void*) im;

	prev_x = prev_y = prev_z = INVALID_COORDINATE;

	w2 = w->child;
	while (w2) 
	{
		if ('<'==*w2->str && w2->tag_number != TAG_NONE) 
		{
			Coordinate x0,y0,z0,x1,y1,z1;
			int pixel = -1;
			Coordinate color;
			String vectors;
			String colorstr;

			vectors = word_fetch_tag_var_value (w2,"v");
			if (!vectors)
				vectors = word_fetch_tag_var_value (w2,"vectors");

			colorstr = word_fetch_tag_var_value (w2,"c");
			if (!colorstr)
				colorstr = word_fetch_tag_var_value (w2,"color");

			if (colorstr) {
				color=color_lookup (colorstr);
				if (color != INVALID_COLOR) {
					pixel = dpy->rgb_to_pixel (color);
				}
				if (debug_mode && logfile) {
					fprintf (logfile, "MODEL_TAG: tag %s will use color %s, which is pixel %d\n",
						w->segments->str, colorstr, pixel);
					fflush(logfile);
				}
			}
			if (pixel < 0) 
				pixel = dpy->fg;

			switch (w2->tag_number) {
			case TAG_PIXEL: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-pixel vector=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_three_coords (vectors,&x0,&y0,&z0)) 
				{
					if (angle_x || angle_y || angle_z)
						three_dimension_transform (&x0, &y0, &z0, 
						   angle_x, angle_y, angle_z);
					x0 -= ix0;
					y0 -= iy0;
					image_draw_pixel (im,x0,y0,pixel);
					prev_x = x0;
					prev_y = y0;
					//prev_z = z0;
				}
				break;
			case TAG_LINEREL: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-line-relative vector=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_three_coords (vectors,&x0,&y0,&z0)) 
				{
					if (angle_x || angle_y || angle_z)
						three_dimension_transform (&x0, &y0, &z0, 
						   angle_x, angle_y, angle_z);
					x0 -= ix0;
					y0 -= iy0;
					if (prev_x != INVALID_COORDINATE &&
					    prev_y != INVALID_COORDINATE)
						image_draw_line (im,prev_x,prev_y,x0,y0,pixel);
					prev_x = x0;
					prev_y = y0;
				}
				break;
			case TAG_LINE: 				
				if (debug_mode && logfile) {
					fprintf (logfile, "draw-line vectors=%s\n", vectors);
					fflush(logfile);
				}
				if (vectors && parse_six_coords (vectors,
					&x0,&y0,&z0,
					&x1,&y1,&z1)) 
				{
					if (angle_x || angle_y || angle_z)
						three_dimension_transform (&x0, &y0, &z0, 
						   angle_x, angle_y, angle_z);
					if (angle_x || angle_y || angle_z)
						three_dimension_transform (&x1, &y1, &z1, 
						   angle_x, angle_y, angle_z);
					x0 -= ix0;
					y0 -= iy0;
					x1 -= ix0;
					y1 -= iy0;
					image_draw_line (im,x0,y0,x1,y1,pixel);
					prev_x = x1;
					prev_y = y1;
					prev_z = z1;
				}
				break;

			case TAG_TEXT: 	{
				Style sl;
				String expr, face, sizestr, italic, bold;
				FontSize size;

				style_reset(&sl);

				if (debug_mode && logfile) 
					fprintf (logfile, "draw-text vectors=%s\n", vectors);

				expr= word_fetch_tag_var_value (w2,"expr");
				face= word_fetch_tag_var_value (w2,"face");
				sizestr= word_fetch_tag_var_value (w2,"size");
				if (sizestr) 
					size=atoi(sizestr);
				else 
					size=INVALID_FONTSIZE;
				italic= word_fetch_tag_var_value (w2,"italic");
				bold= word_fetch_tag_var_value (w2,"bold");

				if (face) sl.family=my_strdup(face);
				sl.size=size;
				sl.flags= (italic?ATTR_ITALIC:0) |
				          (bold?ATTR_BOLD:0);

				/* XX - Need to process inline Style here
				 */

				if (debug_mode && logfile) {
					fprintf(logfile,"MODEL: text ");
					style_dump(logfile,&sl);
					fprintf(logfile,"\n");
				}

				style_fix_font (dpy, &sl);

				if (vectors && parse_three_coords (vectors,&x0,&y0,&z0)) 
				{
					if (angle_x || angle_y || angle_z)
						three_dimension_transform (&x0, &y0, &z0, 
						   angle_x, angle_y, angle_z);

					x0 -= ix0;
					y0 -= iy0;
					image_draw_text (im,
						sl.fontdata,
						sl.flags & ATTR_BOLD,
						sl.flags & ATTR_ITALIC,
						expr,x0,y0,pixel);
					prev_x = x0;
					prev_y = y0;
				}

				if (face) my_free(face);
				break;
			}

			} /* switch */
		}
		w2=w2->next;
	}

	group_append_word (dpy, current_group, w);

	return retval;
}



int
process_ol_ul_tag (DisplayDevice *dpy, Group *current_group, Word *w, 
                int list_type) 
{
	int retval=TRUE;
	Group *subgroup;
	Dimension height;
	Dimension newwidth;
	unsigned doing_ul=FALSE;
	Dimension padding=0;
	unsigned char list_subtype;
	long list_start, list_offset;
	int indent_amount;
	String str;

	ENTER;

	if (w->tag_number == TAG_UL) {
		doing_ul=TRUE;
	} 

	group_end_line (dpy, current_group);

	if (!current_group->beginning_new_paragraph) {
		padding = 7;
	} else {
		padding = 0;
	}

	if (doing_ul) {
		str= word_fetch_tag_var_value (w,"type");
		list_start = list_offset = 0;

		if (str) {
			int ch=tolower(*str);
			list_subtype = UL_TYPE_DISC;
			switch(ch) {
			case 'd': list_subtype=UL_TYPE_DISC; break;
			case 's': list_subtype=UL_TYPE_SQUARE; break;
			case 'c': list_subtype=UL_TYPE_CIRCLE; break;
			}
		}

	} else {
		String str2= word_fetch_tag_var_value (w,"offset");
		String str3= word_fetch_tag_var_value (w,"type");

		str= word_fetch_tag_var_value (w,"start");

		list_subtype=OL_TYPE_ARABIC;

		if (str)
			list_start = atoi(str);
		else
			list_start = 1;
		if (str2)
			list_offset = atoi(str2);
		else
			list_offset = 1;

		if (str3) {
			char first_char = *str3;

			switch (first_char) {
			case '1':  /* arabic */
				break;
			case 'A':  /* uppercase */
				list_subtype= OL_TYPE_UPPER;
				if (list_start<1) list_start=1;
				else if (list_start>26) list_start=26;
				break;
			case 'a':  /* lowercase */
				list_subtype= OL_TYPE_LOWER;
				if (list_start<1) list_start=1;
				else if (list_start>26) list_start=26;
				break;
			case 'I':  /* roman numerals e.g. XVII */
				list_subtype= OL_TYPE_ROMAN_UPPER;
				if (list_start<1) list_start=1;
				else if (list_start>25) list_start=25;
				break;
			case 'i':  /* roman numerals e.g. XVII */
				list_subtype= OL_TYPE_ROMAN_LOWER;
				if (list_start<1) list_start=1;
				else if (list_start>25) list_start=25;
				break;
			}
		}

		if (list_subtype==OL_TYPE_ARABIC) {
			if (list_start<-99999) list_start = -99999;
			else if (list_start>99999) list_start=99999;
		}
	}

	subgroup = group_new (dpy, current_group, 
		0,
		current_group->ystart + padding,  /* pre-paragraph spacing */
		current_group->max_width - current_group->x,
		current_group->max_height);

	w->subgroup = (void*) subgroup;

	str = word_fetch_tag_var_value (w,"bgcolor");
	if (str) {
		subgroup->bgcolor = color_lookup (str);
	}

	subgroup->list_subtype = list_subtype;

fprintf(logfile, "tag=%s, list_subtype = %d\n", w->str,subgroup->list_subtype);

	subgroup->list_start = list_start;
	subgroup->list_offset = list_offset;
	subgroup->list_value = list_start;
	subgroup->list_type = doing_ul ? LIST_DOTTED : LIST_NUMBERED;

	indent_amount = dpy->list_indentation_space;
	if (!doing_ul && 
	    (list_subtype==OL_TYPE_ROMAN_UPPER || 
	     list_subtype==OL_TYPE_ROMAN_LOWER)) {
		short tmp = indent_amount;
		indent_amount *= 2;
		indent_amount += tmp/2;
	}
	subgroup->list_indent = indent_amount;

	subgroup->beginning_new_paragraph = TRUE;

	/* No support yet for list items which
	 * use pictures or text expressions instead of
	 * dots/numbers.
	 */

	if (w->child) {
		if (!group_geometry_core (dpy, subgroup, w->child, 
		  subgroup->list_type))
			retval=FALSE;
	}
	group_end_line (dpy, subgroup);

	height = subgroup->height;
	height += padding;

	w->min_width = w->width = subgroup->width;
	w->ascent = height;
	w->descent = 0;

	/* Now add the post-list padding */
	height += padding;

fprintf(logfile,"adding %s, width=%ld, ascent=%ld\n", w->str,w->width,w->ascent);

	group_append_word (dpy, current_group, w);
	group_end_line (dpy, current_group);

	current_group->beginning_new_paragraph = FALSE;
	current_group->insert_paragraph_space = TRUE;

#if 0
	current_group->ystart += height;
	current_group->height += height;

	newwidth = subgroup->width + current_group->xstart;

	if (newwidth > current_group->width)
	current_group->width = newwidth;
#endif

	return retval;
}

int
process_dl_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	ENTER;

	group_end_line_make_line(dpy, current_group);

	group_append_word (dpy, current_group, w);

	if (w->child) {
		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;
	}

	group_end_line (dpy, current_group);
	return retval;
}

int
process_dt_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	ENTER;

/* XX -- Why is there a LIST_DEFINITION reference down in LI?
 */
	group_end_line_make_line(dpy, current_group);
	/* XXX need to terminate current line here */

	group_append_word (dpy, current_group, w);

	if (w->child) {
		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;
	}

	return retval;
}

int
process_dd_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;
	
	ENTER;

	group_end_line_make_line(dpy, current_group);
	/* XXX need to terminate current line here */
	group_end_line_make_line(dpy, current_group);

	group_append_word (dpy, current_group, w);

	if (w->child) {
		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;
	}

	group_end_line_make_line(dpy, current_group);
	return retval;
}

int
process_li_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	Group *subgroup;
	Dimension height;
	Dimension newwidth;
	Coordinate startx;
	long ch;
	Char str[40];
	String str2;
	long value;
	int list_subtype;

	ENTER;

	group_end_line (dpy, current_group);
#if 0
	group_end_line_make_line (dpy, current_group);
#endif

	list_subtype = current_group->list_subtype;

	str2= word_fetch_tag_var_value (w,"type");
	if (str2) {
		int ch=tolower(*str2);
		switch(ch) {
		case 'd': list_subtype=UL_TYPE_DISC; break;
		case 's': list_subtype=UL_TYPE_SQUARE; break;
		case 'c': list_subtype=UL_TYPE_CIRCLE; break;
		case '1': list_subtype=OL_TYPE_ARABIC; break;
		case 'i': list_subtype=OL_TYPE_ROMAN_LOWER; break;
		case 'I': list_subtype=OL_TYPE_ROMAN_UPPER; break;
		case 'a': list_subtype=OL_TYPE_LOWER; break;
		case 'A': list_subtype=OL_TYPE_UPPER; break;
		}
	}

	str2= word_fetch_tag_var_value (w,"value");
	if (str2)
		value = atoi(str2);
	else
		value = current_group->list_value;

	/* We create the list item here, not during
	 * redraw. The problem is that partial redraws
	 * may exclude list items, so if we wanted to
	 * determine the list item's number e.g. 5th
	 * in the list, we would have to search for
	 * all LI tags in the Group. Not efficient!
	 */
	switch (list_type) 
	{
	case LIST_NONE:
	case LIST_NUMBERED:
	case LIST_DOTTED:
		switch (list_subtype) {
		case UL_TYPE_DISC:
			strcpy (str, "&DISC;");
			break;
		case UL_TYPE_SQUARE:
			strcpy (str, "&SQUARE;");
			break;
		case UL_TYPE_CIRCLE:
			strcpy (str, "&CIRCLE;");
			break;
		case OL_TYPE_UPPER:
			ch=current_group->list_value+'A'-1;
			if(ch<'A') ch='A'; if(ch>'Z') ch='Z';
			str[0]=ch; str[1]='.'; str[2]=' '; str[3]=0;
			break;
		case OL_TYPE_LOWER:
			ch=current_group->list_value+'a'-1;
			if(ch<'a') ch='a'; if(ch>'z') ch='z';
			str[0]=ch; str[1]='.'; str[2]=' '; str[3]=0;
			break;
		case OL_TYPE_ARABIC:
			sprintf (str,"%ld. ", current_group->list_value);
			break;
		case OL_TYPE_ROMAN_UPPER: 
			generate_roman_numeral (current_group->list_value, str, 
				100, TRUE);
			strcat (str,". ");
			break;
		case OL_TYPE_ROMAN_LOWER: 
			generate_roman_numeral (current_group->list_value, str, 
				100, FALSE);
			strcat (str,". ");
			break;
		}
	  	break;

	case LIST_DEFINITIONS:
		str[0]=0;
	  	w->width=0;
	  	w->ascent=0;
	  	w->descent=0;
	  	break;
	}

	current_group->list_value += current_group->list_offset; 

	w->tag.ptr = (void*) my_strdup (str);

	if (!strcmp("&DISC;",str) ||
	    !strcmp("&SQUARE;",str) ||
	    !strcmp("&CIRCLE;",str))
	{
		w->width=8;
		w->ascent=8;
		w->descent=0;
	}
	else
		group_text_dims (dpy, current_group, str, &w->width, &w->ascent,
			&w->descent);

	/* No support yet for list items which
	 * use pictures or expressions instead of
	 * dots/numbers.
	 */

	if (current_group->x > current_group->list_indent)
		startx = current_group->x;
	else
		startx = current_group->list_indent;

	subgroup = group_new (dpy, current_group, 
		startx,
		current_group->ystart,  /* pre-paragraph spacing */
		current_group->max_width-startx,
		current_group->max_height);

	w->subgroup = (void*) subgroup;

	subgroup->beginning_new_paragraph = FALSE;

	str2 = word_fetch_tag_var_value (w,"bgcolor");
	if (str2) {
		subgroup->bgcolor = color_lookup (str2);
	}

#if 0
printf ("reached call to geometry\n");
fflush(stdout);
#endif

	if (w->child) {
		if (!group_geometry (dpy, subgroup, w->child))
			retval=FALSE;
	}

#if 0
printf ("reached call to endline\n");
fflush(stdout);
#endif

	height = subgroup->height;

	/* The LI trick: the subgroup refers to a 
	 * different set of data than the counter/bullet.
	 */
	w->width = w->min_width = subgroup->width + startx;
	w->ascent = subgroup->height;
	w->descent = 0;
	group_append_word (dpy, current_group, w);

	current_group->beginning_new_paragraph = FALSE;
	current_group->insert_paragraph_space = FALSE;

#if 0
	/* Subtract the height of the line which has the
	 * dot, number, image etc on it, since the subgroup
	 * is hanging down from that line -- unless 
	 * of course it is an empty LI tag.
	 */
	if (height) {
		height -= (current_group->work_line->ascent +
			   current_group->work_line->descent);
	}

	current_group->ystart += height;
	current_group->height += height;

	newwidth = subgroup->x0 + subgroup->width + current_group->xstart;

	if (newwidth > current_group->width)
		current_group->width = newwidth;

	// group_end_line (dpy, current_group);
#endif

#if 0
printf ("exiting li_tag\n");
fflush(stdout);
#endif

	return retval;
}


/*-------Tables----------------------------------------------------------*/
int
process_table_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	Table *tbl;
	Dimension max_width;
	String str;

	int retval=TRUE;
	/* unsigned char align=ALIGN_LEFT; */
	Group *subgroup;
	Dimension height;
	Dimension padding;
	Dimension newwidth;

	ENTER;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	tbl = table_new ();
	w->tag.table = tbl;

	max_width = current_group->max_width;

	str = word_fetch_tag_var_value (w,"width");
	if (str && isdigit(*str)) {
		unsigned long width = atoi(str);
		if (strchr(str,'%')) {
			width = (width*max_width)/100;
		} else if (dpy->width<320) {
			width /= 8;
		}
		tbl->overall.desired_width=width;
	} else {
		/* Set up default table width */
		/* no default */
	}
	
	str = word_fetch_tag_var_value (w,"cellpadding");
	if (str && isdigit(*str)) {
		unsigned long width = atoi(str);
		if (dpy->width<320) {
			width /= 8;
		}
		tbl->cell_padding = width;
	} else {
		/* Set up default table cell padding */
		tbl->cell_padding = dpy->table_cell_padding;
	}
	
	str = word_fetch_tag_var_value (w,"cellspacing");
	if (str && isdigit(*str)) {
		unsigned long width = atoi(str);
		if (dpy->width<320) {
			width /= 8;
		}
		tbl->cell_spacing=width;
	} else {
		/* Set up default table cell spacing */
		tbl->cell_spacing = dpy->table_cell_spacing;
	}
	
	str = word_fetch_tag_var_value (w,"border");
	if (str && isdigit(*str)) {
		int size=atoi (str);
		if (size) {
			if (dpy->width < 320) {
				size /= 4;
				if (!size) size=1;
			}
		}
		tbl->border_size = size;
	} else {
		/* Set up default border width */
		tbl->border_size = dpy->table_border_size;
	}
	
	subgroup = group_new (dpy, current_group, 
		0, 0,
		current_group->max_width,
		MAX_COORD);

	w->subgroup = (void*) subgroup;

	subgroup->beginning_new_paragraph = FALSE;

	str = word_fetch_tag_var_value (w,"bgcolor");
	if (str) {
		subgroup->bgcolor = color_lookup (str);
	}

	/* Get the table's geometry, but if it will not
	 * fit, let's get that error condition back
	 * back to main() as quickly as possible.
	 */
	if (!table_geometry (dpy, subgroup, w->child, tbl)) {
		table_free(tbl);
		w->tag.datum = 0;
		retval=FALSE;
	}

	w->width = w->min_width = tbl->overall.optimal_width;
	w->ascent =    tbl->overall.optimal_height;
	w->descent =   0;

#if 0
fprintf (logfile, "table has dimensions %ld %ld\n", w->width, w->ascent);
fflush(logfile);
#endif

	group_append_word (dpy, current_group, w);

	/* Note, we do not yet have coordinates for the table
	 * since the line has not yet been aligned etc.
	 */

	tbl->word = w;

	return retval;
}

/*-------Font attributes----------------------------------------------------*/
int
process_attr_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	Style *sl;
	int retval=TRUE;

	ENTER;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	sl = style_new();

	if (w->child) {

		switch (w->tag_number) {
		case TAG_B:
			sl->flags = ATTR_BOLD;
			break;
		case TAG_I:
			sl->flags = ATTR_ITALIC;
			break;
		case TAG_U:
			sl->flags = ATTR_UNDERLINE;
			break;
		case TAG_OVER:
			sl->flags = ATTR_OVERBAR;
			break;
		case TAG_STRIKE:
			sl->flags = ATTR_STRIKE;
			break;
		case TAG_EM:
		case TAG_DFN:
		case TAG_VAR:
		case TAG_ACRONYM:
			sl->flags = ATTR_ITALIC;
			break;
		case TAG_ABBR:
		case TAG_KBD:
		case TAG_STRONG:
			sl->flags = ATTR_BOLD; 
			break;
		case TAG_SUP:
			sl->flags = ATTR_SUPER; 
			break;
		case TAG_SUB:
			sl->flags = ATTR_SUB;
			break;
		}

		group_push_style (dpy, current_group, sl);

		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;

		group_drop_style (current_group);
	}

	return retval;
}


int
process_monospace_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) 
{
	char ch;
	int retval=TRUE;
	Style *sl;

	ENTER;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	sl = style_new();

	if (w->child) {
		ch = *w->segments->str;

		switch (ch) {
		case 'k': /* <KBD> */
			sl->flags = ATTR_BOLD; /* bold+courier */
		case 's': /* <SAMP> */
		case 'c': /* <CODE> */ 
		case 't': /* <TT> */
			sl->family = my_strdup("Courier");
			break;
		}

		group_push_style (dpy, current_group, sl);

		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval = FALSE;

		group_drop_style (current_group);
	}

	return retval;
}

int
process_html_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	ENTER;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	/* For now, let's not bother with a subgroup.
	 */
	if (w->child) {
		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;
	}

	return retval;
}


/*-------Header Section-----------------------------------------------------*/
int
process_header_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	/* Ignored at present */

	ENTER;

	return TRUE;
}

int
process_title_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	/* Ignored at present */

	ENTER;

	return TRUE;
}

int
process_body_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;
	Group *subgroup;
	String str;
	Dimension height;

	ENTER;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	/* In case there are multiple BODY tags for some reason.
	 */
	w->width = 0;
	w->ascent = 0;
	w->descent = 0;
	group_end_line (dpy, current_group);

	subgroup = group_new (dpy, current_group, 
		0,
		current_group->ystart,  /* pre-paragraph spacing */
		current_group->max_width-(current_group->x0+current_group->x),
		current_group->max_height);

	w->subgroup = (void*) subgroup;

	str = word_fetch_tag_var_value (w,"bgcolor");
	if (str) {
		subgroup->bgcolor = color_lookup (str);
	}

	subgroup->beginning_new_paragraph = FALSE;

	if (w->child) {
		if (!group_geometry (dpy, subgroup, w->child))
			retval=FALSE;
	}

	w->width = w->min_width = subgroup->width;
	w->ascent = subgroup->height;
	w->descent = 0;
	group_append_word (dpy, current_group, w);
	group_end_line (dpy, current_group);

	current_group->beginning_new_paragraph = FALSE;
	current_group->insert_paragraph_space = FALSE;
#if 0
	height = subgroup->height;
	current_group->ystart += height;
	current_group->height += height;

	{
		int newwidth = subgroup->width + current_group->xstart;

		if (newwidth > current_group->width)
			current_group->width = newwidth;
	}
#endif

	return retval;
}

int
process_heading_tag (DisplayDevice *dpy, Group *current_group, Word *w, 
  int list_type) 
{
	char which;
	int retval=TRUE;
	Group *subgroup;
	Dimension height;
	Dimension padding;
	String str;

	ENTER;

	group_end_line (dpy, current_group);

	if (!current_group->beginning_new_paragraph) {
		padding = 7;
	} else {
		padding = 0;
	}

	subgroup = group_new (dpy, current_group, 
		current_group->x,
		current_group->ystart+ padding ,  /* pre-paragraph spacing */
		current_group->max_width-(current_group->x0+current_group->x),
		current_group->max_height);

	w->subgroup = (void*) subgroup;

	str = word_fetch_tag_var_value (w,"bgcolor");
	if (str) {
		subgroup->bgcolor = color_lookup (str);
	}

	subgroup->beginning_new_paragraph = FALSE;

	if (w->child) {
		Style *sl = style_new();
		sl->name = my_strdup (w->segments->str);

		which = w->segments->str[1];

		switch (which) {
		case '1': sl->size=24; break;
		case '2': sl->size=18; break;
		case '3': sl->size=14; break;
		case '4': sl->size=12; break;
		case '5': sl->size=10; break;
		case '6': sl->size=8; break;
		}

		sl->flags = ATTR_PLAIN | ATTR_BOLD;

		group_push_style (dpy, subgroup, sl);

		if (!group_geometry_core (dpy, subgroup, w->child, LIST_NONE))
			retval=FALSE;

		group_drop_style (subgroup);
		group_end_line (dpy, subgroup);
	}

	height = subgroup->height;
	height += padding;

	w->width = w->min_width = subgroup->width;
	w->ascent = height;
	w->descent = 0;

	group_append_word (dpy, current_group, w);
	group_end_line (dpy, current_group);

	current_group->beginning_new_paragraph = FALSE;
	current_group->insert_paragraph_space = TRUE;

#if 0
	current_group->ystart += height;
	current_group->height += height;

	{
		int newwidth = subgroup->width + current_group->xstart;

		if (newwidth > current_group->width)
			current_group->width = newwidth;
	}
#endif

	return retval;
}


int
process_big_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	ENTER;

	if (w->child) {
		Style *sl = style_new();

		group_get_total_style (dpy, current_group, sl);
		sl->size=determine_bigger_standard_size (sl->size);
		group_push_style (dpy, current_group, sl);

		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;

		group_drop_style (current_group);
	}

	return retval;
}

int
process_small_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) {
	int retval=TRUE;

	ENTER;

	if (w->child) {
		Style *sl = style_new();

		group_get_total_style (dpy, current_group, sl);
		sl->size=determine_smaller_standard_size (sl->size);
		group_push_style (dpy, current_group, sl);

		if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
			retval=FALSE;

		group_drop_style (current_group);
	}

	return retval;
}


/*-------Font foreground colors etc-----------------------------------------*/
int
process_font_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type) 
{
	unsigned long val;
	Style *sl;
	char *str;
	int retval=TRUE;

	ENTER;

	if (!w->child) 
		return TRUE;

	sl=style_new();

	/* XX - common color names could be redad in from rgb.txt &
	 * hashed for speed
	 */

	str = word_fetch_tag_var_value (w,"color");
	if (str) {
		if (*str=='#' && 1==sscanf (str+1, "%6lx", &val)) {
			sl->fg = val;
		} else {
			sl->fg = color_lookup (str);
		} 
	}

	/* Bgcolor is probably not valid HTML, but I like it.
	 */
	str = word_fetch_tag_var_value (w,"bgcolor");
	if (str) {
		sl->bg = color_lookup (str);
	}

	str = word_fetch_tag_var_value (w,"size");
	if (str) {
		FontSize size = group_get_fontsize (current_group);
		if (*str=='+') {
			int i = atoi(str+1);
			while(i--) {
				size = determine_bigger_standard_size (size);
			}
			sl->size = size;
		} 
		else if (*str=='-') {
			int i = atoi(str+1);
			while(i--) {
				size = determine_smaller_standard_size (size);
			}
			sl->size = size;
		}
		else if (isdigit(*str)) {
			char *s = strstr(str,"pt");
			int i = atoi(str);
			if (i && s && i<256)
				sl->size = i;
		}
	}

	str = word_fetch_tag_var_value (w,"face");
	if (str)
		sl->family = my_strdup (str);

/* printf ("NEW FONT SIZE %d\n",sl->size); */
	group_push_style (dpy, current_group, sl);

	if (!group_geometry_core (dpy, current_group, w->child, LIST_NONE))
		retval=FALSE;

	group_drop_style (current_group);

/* printf ("process_font_tag returning %d\n", retval); */
	return retval;
}

typedef struct {
	char *name;
	int (*func)(DisplayDevice*,Group*,Word*,int);
	unsigned char number;
}
TagHashItem;



static TagHashItem tags_a[] = {
	{"a",process_a_tag, TAG_A}, 
	{"author",process_attr_tag, TAG_AUTHOR},
	{"abbr",process_attr_tag, TAG_ABBR}, 
	{"acronym",process_attr_tag, TAG_ACRONYM},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_b[] = {
	{"b", process_attr_tag, TAG_B}, 
	{"br",process_break_tag, TAG_BR},
	{"big",process_big_tag, TAG_BIG}, 
	{"body",process_body_tag, TAG_BODY},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_c[] = {
	{"cite",process_attr_tag, TAG_CITE}, 
	{"code",process_monospace_tag, TAG_CODE},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_d[] = {
	{"dd",process_dd_tag, TAG_DD}, 
	{"dfn",process_attr_tag, TAG_DFN}, 
	{"dl",process_dl_tag, TAG_DL}, 
	{"dt",process_dt_tag, TAG_DT}, 
	{"draw",process_draw_tag, TAG_DRAW}, 
	{"date",NULL, TAG_DATE},
	{"del", process_ignore_tag, TAG_DEL},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_e[] = {
	{"em",process_attr_tag, TAG_EM},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_f[] = {
	{"font", process_font_tag, TAG_FONT},
	{"filledrect",NULL, TAG_FILLEDRECT},
	{"filledoval",NULL, TAG_FILLEDOVAL},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_g[] = {{NULL,NULL,TAG_NONE}};

static TagHashItem tags_h[] = {
	{"h1",process_heading_tag, TAG_H1}, 
	{"h2",process_heading_tag, TAG_H2}, 
	{"h3",process_heading_tag, TAG_H3}, 
	{"hr",process_hr_tag, TAG_HR}, 
	{"h6",process_heading_tag, TAG_H6}, 
	{"h4",process_heading_tag, TAG_H4}, 
	{"h5",process_heading_tag, TAG_H5}, 
	{"html",process_html_tag, TAG_HTML}, 
	{"head",process_header_tag, TAG_HEAD}, 
	{"h~",NULL, TAG_HLINE},
	{"hline",NULL, TAG_HLINE},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_i[] = {
	{"i", process_attr_tag, TAG_I}, 
	{"img",process_img_tag, TAG_IMG},
	{"ins", process_passthrough_tag, TAG_INS},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_j[] = {{NULL,NULL,TAG_NONE}};

static TagHashItem tags_k[] = {
	{"kbd",process_monospace_tag, TAG_KBD},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_l[] = {	
	{"li",process_li_tag, TAG_LI},
	{"l~",NULL, TAG_LINE},
	{"lr~",NULL, TAG_LINEREL},
	{"line",NULL, TAG_LINE},
	{"linerel",NULL, TAG_LINEREL},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_m[] = {
	{"model",process_model_tag, TAG_MODEL}, 
	{"menu", process_ol_ul_tag, TAG_UL},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_n[] = { 
	{"noscript",process_passthrough_tag, TAG_NOSCRIPT}, 
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_o[] = {
	{"ol",process_ol_ul_tag, TAG_OL},
	{"o~",NULL, TAG_OVAL},
	{"oval",NULL, TAG_OVAL},
	{"over",process_attr_tag, TAG_OVER},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_p[] = {
	{"p", process_para_tag, TAG_P}, 
	{"pre", NULL, TAG_PRE},
	{"p~", NULL, TAG_PIXEL},
	{"pixel", NULL, TAG_PIXEL},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_q[] = {
	{"q", process_passthrough_tag, TAG_Q}, 
        {NULL,NULL,TAG_NONE}
};

static TagHashItem tags_r[] = {
	{"r~", NULL, TAG_RECT},
	{"rect", NULL, TAG_RECT},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_s[] = {
	{"small",process_small_tag, TAG_SMALL}, 
	{"span",NULL, TAG_SPAN}, 
	{"sup",process_attr_tag, TAG_SUP},
	{"sub",process_attr_tag, TAG_SUB},
	{"s",process_attr_tag, TAG_STRIKE}, 
	{"strike",process_attr_tag, TAG_STRIKE}, 
	{"strong",process_attr_tag, TAG_STRONG}, 
	{"script",process_ignore_tag, TAG_SCRIPT}, 
	{"samp",process_monospace_tag, TAG_SAMP},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_t[] = {
	{"table", process_table_tag, TAG_TABLE}, 
	{"th",NULL, TAG_TH},
	{"td",NULL, TAG_TD},
	{"tr",NULL, TAG_TR},
	{"tt",process_monospace_tag, TAG_TT},
	{"t~",NULL, TAG_TEXT},
	{"text",NULL, TAG_TEXT},
	{"title",process_title_tag, TAG_TITLE},
	{"tbody",NULL, TAG_TBODY},
	{"thead",NULL, TAG_THEAD},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_u[] = {
	{"u", process_attr_tag, TAG_U}, 
	{"ul",process_ol_ul_tag, TAG_UL},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_v[] = {	
	{"var",process_attr_tag, TAG_VAR},
	{"v~",NULL, TAG_VLINE},
	{"vline",NULL, TAG_VLINE},
	{NULL,NULL,TAG_NONE}
};

static TagHashItem tags_w[] = {{NULL,NULL,TAG_NONE}};
static TagHashItem tags_x[] = {{NULL,NULL,TAG_NONE}};
static TagHashItem tags_y[] = {{NULL,NULL,TAG_NONE}};
static TagHashItem tags_z[] = {{NULL,NULL,TAG_NONE}};


#if 0
static int tags_counts[] = {
#define _TC(XX) (sizeof(XX)/sizeof(TagHashItem))
	_TC(tags_a), _TC(tags_b), _TC(tags_c), _TC(tags_d), _TC(tags_e), _TC(tags_f),
	_TC(tags_g), _TC(tags_h), _TC(tags_i), _TC(tags_j), _TC(tags_k), _TC(tags_l),
	_TC(tags_m), _TC(tags_n), _TC(tags_o), _TC(tags_p), _TC(tags_q), _TC(tags_r),
	_TC(tags_s), _TC(tags_t), _TC(tags_u), _TC(tags_v), _TC(tags_w), _TC(tags_x),
	_TC(tags_y), _TC(tags_z)
};
#endif

static
TagHashItem *tag_hash[] = {
	tags_a, tags_b, tags_c, tags_d, tags_e, tags_f, tags_g, tags_h, tags_i,
	tags_j, tags_k, tags_l, tags_m, tags_n, tags_o, tags_p, tags_q, tags_r, tags_s,
	tags_t, tags_u, tags_v, tags_w, tags_x, tags_y, tags_z,
};


static int (*tag_array[TAG_MAXVALUE])(DisplayDevice*,Group*,Word*,int);


enum { EXEC_FAIL, EXEC_SUCCESS, EXEC_UNKNOWN };


/* Hash data is put into an array for when tags are executed.
 * The hash speeds up parsing, the array speeds up execution.
 */
void
tagarray_init ()
{
	int i,j;
	TagHashItem *th;

	for (i=0; i<TAG_MAXVALUE; i++)
		tag_array [i] = NULL;

	for (i=0; i<26; i++) {
		j=0;
		th = tag_hash[i];
		while (th[j].name) {
			tag_array [th[j].number] = th[j].func;
			j++;
		}
	}
}


int
taghash_lookup_number (String tag)
{
	TagHashItem *th;
	String name;
	int i;

	CHECK_PARAM_NOT_NULL (tag);

	i= *tag - 'a';
	if (i< 0 || i>= 26)
		return TAG_NONE;

	th = tag_hash [i];
	i=0;
	while ((name = th[i].name)) {
		if (!strcmp (tag, name)) {
			return th[i].number;
		}
		i++;
	}

	return TAG_NONE;
}



int
execute_tag (DisplayDevice *dpy, Group *current_group, Word *w, int list_type)
{
	int i;
	int ch;
	char *str;
	int rv=TRUE;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	if (!w->segments)
		error_handler ("tag Word lacks Segments");

	ch = *w->str;
	if (ch != '<')
		error_handler ("execute_tag called for non-tag");

	w->width = w->ascent = w->descent = 0;

	/* If the HTML coder has been sloppy, we may find
	 * that the parser has passed us a closing tag that
	 * it could not match to an opening tag.
	 */
	if (w->str[1] == '/')
		return TRUE;

	/* Abort if the tag does not begin with a lower case letter.
	 * The parser makes all tags lower case.
	 */
	str = w->segments->str;
	if (!str)
		error_handler ("tag Segments lacks string");
	ch = *str;
	if (ch<'a' || ch>'z') {
		warning_handler ("tag does not begin wtih a lower-case letter");
		return TRUE;
	}

/* printf ("EXECUTING TAG %s\n", w->str); */

	/* Now we must call the handler routine (if any).
	 * If there is no handler routine, just call the children (if any).
	 */
	if ((i = w->tag_number))
	{
		if (tag_array[i]) 
			return tag_array[i] (dpy, current_group, w, list_type);
		else {
			if (w->child)
				return group_geometry_core (dpy, current_group, w->child, 
			  		LIST_NONE);
			else
				return TRUE;
		}
	}

	/* We reach here when the tag is completely unknown to us.
	 * Process its child if there are any, else ignore it.
	 */
	if (!w->child) {
		if (debug_mode && logfile) 
			printf ("Ignoring childless unknown tag %s...\n", w->str);
		rv=TRUE;
	} else {
		if (debug_mode && logfile) 
			printf ("Processing children of unknown tag %s...\n", w->str);

		rv = group_geometry_core (dpy, current_group, w->child, 
		  LIST_NONE);
		if (!rv)
			printf ("Exec: execution of children of unfamiliar tag failed\n");
	}
/* printf ("RETURNING FROM EXECUTING unknown TAG %s\n", w->str); */
	return rv;
}



/*-----------------------Group Geometry Functionality-----------------------*/


static int
group_geometry_core (DisplayDevice *dpy, Group *current_group, Word *w, int list_type)
{
	char *s;
	int retval=TRUE;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(current_group);
	CHECK_PARAM_NOT_NULL(w);

	if (debug_mode && logfile) {
		fprintf (logfile, "Upon entering core, word='%s'\n", w->str);
		fflush(logfile);
	}

	/* Ignore starting whitespace Word.
	 */
	if (w->str[0] == ' ') {
		w=w->next;
#if 0
printf ("skipping first whitespace word, next word is '%s'\n", w?(char*)w->str:"no word");
#endif
	}

	while(w) {

		/**********************************************************
		 * STEP 0: Get string and basic dimensions
		 **********************************************************/

		s=w->str;
		if(!s) error_handler ("Word without string");
		w->width = 0;
		w->ascent = w->descent = 0;

		if (*s=='<') {
			if (!execute_tag (dpy,current_group,w,list_type)) {
				retval=FALSE;
				break;
			}
		} else 
		if(*s== '&') {
			Dimension wd, ascent, descent;
			int ch = dpy->convert_special_expr(s);
			char tmp[2];
			tmp [0]=ch;
			tmp [1]=0;
			w->tag.datum = ch;
	  		group_text_dims (dpy, current_group, tmp, &wd, &ascent,
				&descent);
			w->width = wd;
			w->min_width = wd;
			w->ascent = ascent;
			w->descent = descent;
			group_append_word (dpy, current_group, w);
		} else {
			Dimension wd, ascent, descent;
	  		group_text_dims (dpy, current_group, w->str, &wd, 
				&ascent,
				&descent);
/* printf ("DIMS group text dims returned %d by %d for %s\n",wd,ht,s); */
			if (wd > current_group->max_width) {
				/* Need to split object into >1 pieces
				 * i.e. create more Word objs.
				 * Example: a 10000 char word containing
				 * one long hex dump. Current system
				 * cannot deal with it.
				 */
				w->width = wd;
				w->min_width = wd;
				w->ascent = ascent;
				w->descent = descent;
			} else {
				w->width = wd;
				w->min_width = wd;
				w->ascent = ascent;
				w->descent = descent;
			}
			group_append_word (dpy, current_group, w);
		}

		w=w->next;

		/* If the last Word is whitespace,
		 * then ignore it.
		 */
		if (w && !w->next && w->str[0]==' ') break;
	}

	return retval;
}



/*------------------------------------------------------------------------
 * Text positioning routine, including execution of tags.
 */

int
group_geometry (DisplayDevice *dpy, Group *current_group, Word *w)
{ 
	int success;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(w);

	success = 
	  group_geometry_core (dpy, current_group, w, LIST_NONE);

	if (success) 
	{
		group_end_line (dpy, current_group);

		if (debug_mode && logfile) {
			fprintf (logfile, "%lu lines are taking up %lu rasters\n",
				current_group->total_lines,
				current_group->height);

			group_dump (logfile, current_group);
			fflush(logfile);
		}
	}

	return success;
}



int
group_identify_target_core (Group *g,
	Coordinate x, Coordinate y, /* absolute */
	Coordinate parents_x, Coordinate parents_y,
	SelectionPoint *sp)
{
	Word *w;
	Line *l;
	Coordinate base_x, base_y;

	ENTER;

	CHECK_PARAM_NOT_NULL(g);

	base_x = parents_x + g->x0;
	base_y = parents_y + g->y0;

	if (debug_mode && logfile) {
		fprintf (logfile, "IDENTIFY_TARGET: absolute x=%ld absolute y=%ld\n",x,y);
		fflush(logfile);
	}
 
	l = g->line_list;

	while (l) {
		Coordinate ly0 = l->y;
		Coordinate ly1 = l->y + l->ascent + l->descent - 1;

		ly0 += base_y;
		ly1 += base_y;

		/* If we are within the area of the target spot,
		 * then let us work on this line.
		 */
		w = l->word_list; 

		while (w) {
			Coordinate wx0, wx1;
			Group *subgroup;

			wx0 = w->x + base_x;
			wx1 = wx0 + w->width-1;

			subgroup = (Group*) w->subgroup;

			if (!subgroup && y >= ly0 && y <= ly1 &&
			    wx0 <= x && wx1 >= x) 
			{
				long so=sizeof (SelectionPoint);
				memset((void*)sp,0,so);

				sp->ax = wx0;
				sp->ay = w->y + base_y;
				sp->group = g;
				sp->line = l;
				sp->word = w;
				sp->ly0 = ly0;
				sp->ly1 = ly1;

				return sp;
			}

			if (subgroup) {
				if (group_identify_target_core (
					subgroup, x, y,
					base_x, base_y,
					sp))
				return TRUE;
			}

			w = w->next_in_line;
		}
		l = l->next;
	}

	return NULL;
}


/*------------------------------------------------------------------------
 * Name:     group_identify_target
 * Purpose:  Goes through Group tree and determines what object is at
 *           at given x,y absolute location. Used when a user has clicked on 
 *           something.
 *----------------------------------------------------------------------*/

int
group_identify_target (Group *g, Coordinate x, Coordinate y,
	SelectionPoint *sp)
{
	return group_identify_target_core (g,x,y,0,0,sp);
}



#if 0
void
group_highlight_word (DisplayDevice *dpy, Word *w)
{
	Coordinate x1, y1;
	x1 = w->x + w->width-1;
	y1 = w->y + w->ascent + w->descent - 1;
	dpy->reset_colors();
	dpy->draw_rectangle (w->x, w->y, x1, y1);
}
#endif



/*------------------------------------------------------------------------
 * Name:     group_write_html
 * Purpose:  Goes through Group tree and writes it out all to an HTML file.
 *----------------------------------------------------------------------*/

void
group_write_html (FILE *f, DisplayDevice *dpy, Group *g)
{
	char *str;
	Line *l;
#if 0
	Coordinate start_y, end_y;
#endif
	Style previous_sl;
	unsigned char have_previous_sl=FALSE;

	CHECK_PARAM_NOT_NULL(f);
	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(g);

	style_reset (&previous_sl);

	while (g) {
		Word *w;
		unsigned char line_has_words=FALSE;

		l = g->line_list;

		while (l) 
		{
			/* If we are within the area to be redrawn,
			 * then let us work on this line.
			 */
			w = l->word_list; 

			while (w) {
				Style *sl = w->style;
				unsigned char flags;

				str = w->str;
				if (!str)
					error_handler ("word lacks string");

				if (w->tag_number == TAG_BR) {
					fprintf (f, "<br>\n");
					w=w->next;
					continue;
				}

				/* Undo the effects of a previous named 
				 * Style.
				 */
				if (sl && have_previous_sl && previous_sl.name) {
					if (!strcmp("h1",previous_sl.name)) {
						fprintf (f, "</h1>\n");
					} else
					if (!strcmp("h2",previous_sl.name)) {
						fprintf (f, "</h2>\n");
					} else
					if (!strcmp("h3",previous_sl.name)) {
						fprintf (f, "</h3>\n");
					} else
					if (!strcmp("h4",previous_sl.name)) {
						fprintf (f, "</h4>\n");
					} else
					if (!strcmp("h5",previous_sl.name)) {
						fprintf (f, "</h5>\n");
					} else
					if (!strcmp("h6",previous_sl.name)) {
						fprintf (f, "</h6>\n");
					} else
					{
fprintf(f,"<!-- unknown named style: %s -->\n", previous_sl.name);
					}
				}

				/* If we have a new Style, and the previous
				 * Style was *not* a named Style, then
				 * we should undo the effects of the previous
				 * Style, piecemeal.
				 */
				if (sl && !sl->name && 
				    !(have_previous_sl && previous_sl.name))
				{
					unsigned short old_attr;
					unsigned short attr;
					old_attr = previous_sl.flags;
					attr = sl->flags;

					if (!(attr & ATTR_BOLD) && (old_attr & ATTR_BOLD))
					  fprintf(f,"</b>");
					if (!(attr & ATTR_UNDERLINE) && (old_attr & ATTR_UNDERLINE))
					  fprintf(f,"</u>");
					if (!(attr & ATTR_ITALIC) && (old_attr & ATTR_ITALIC))
					  fprintf(f,"</i>");
					if (!(attr & ATTR_STRIKE) && (old_attr & ATTR_STRIKE))
					  fprintf(f,"</strike>");
					if (!(attr & ATTR_LINK) && (old_attr & ATTR_LINK))
					  fprintf(f,"</a>");
				}

				/* Apply line breaks etc.
				 */
				flags = w->flags;
#if 0
				if (flags & WORD_HAS_PRECEDING_LINE_BREAK)
					fprintf(f,"<br>");
#endif
				if (flags & WORD_HAS_PRECEDING_PARAGRAPH_BREAK) {
					switch (w->para_align) {
					case ALIGN_RIGHT: 
						fprintf(f,"<p align=right>");
						break;
					case ALIGN_CENTER: 
						fprintf(f,"<p align=center>");
						break;
					case ALIGN_JUSTIFY: 
						fprintf(f,"<p align=justify>");
						break;
					default:
						fprintf(f,"<p>");
					}
				}

				/* Apply a regular Style if the new Style is
				 * not named.
				 */
				if (sl && !sl->name) {
					unsigned short old_attr;
					unsigned short attr;
					old_attr = previous_sl.flags;
					attr = sl->flags;

					if ((attr & ATTR_BOLD) && !(old_attr & ATTR_BOLD))
					  fprintf(f,"<b>");
					if ((attr & ATTR_UNDERLINE) && !(old_attr & ATTR_UNDERLINE))
					  fprintf(f,"<u>");
					if ((attr & ATTR_ITALIC) && !(old_attr & ATTR_ITALIC))
					  fprintf(f,"<i>");
					if ((attr & ATTR_STRIKE) && !(old_attr & ATTR_STRIKE))
					  fprintf(f,"<strike>");
					if ((attr & ATTR_LINK) && !(old_attr & ATTR_LINK))
					  fprintf(f,"<a>");
				}

				/* Apply a named Style.
				 */
				if (sl && sl->name) {
					unsigned short old_attr;
					unsigned short attr;
					old_attr = previous_sl.flags;
					attr = sl->flags;

					if (!strcmp("h1",sl->name)) {
						fprintf (f, "<h1>");
					} else
					if (!strcmp("h2",sl->name)) {
						fprintf (f, "<h2>");
					} else
					if (!strcmp("h3",sl->name)) {
						fprintf (f, "<h3>");
					} else
					if (!strcmp("h4",sl->name)) {
						fprintf (f, "<h4>");
					} else
					if (!strcmp("h5",sl->name)) {
						fprintf (f, "<h5>");
					} else
					if (!strcmp("h6",sl->name)) {
						fprintf (f, "<h6>");
					} else
					if (!strcmp("a",sl->name)) {
						fprintf (f, "<a href=nothing.html>");
					} else
					{
fprintf(f,"<!-- unknown named style: %s -->\n", previous_sl.name);
					}
				}

				/* If we have a Style, save a copy
				 * so that we can undo its effects later.
				 */
				if (sl) {
					style_copy (&previous_sl, sl);
					have_previous_sl = TRUE;
				}
				else
					have_previous_sl = FALSE;

				if ('&' == str[0]) 
				{
					fprintf(f, "%s\n", w->str);
				} 
				else if ('<' != str[0]) 
				{
					if (w->flags & WORD_HAS_FOLLOWING_WHITESPACE)
					  fprintf(f, "%s ", str);
					else
					  fprintf(f, "%s", str);

					line_has_words=TRUE;
				} else {
					switch (w->tag_number) 
					{

					case TAG_TABLE: {
						Table *t = w->tag.table;
						if (t) {
							/* table_write_html (dpy,t,w); */
							printf ("<!--table-->");
						}
						break;
					}

					case TAG_LI: {
						fprintf(f,"<li>");
						break;
					}

					case TAG_IMG: {
						fprintf(f,"<!--image-->\n");
						break;
					}

					case TAG_HR: {
						fprintf(f,"<hr>");
						break;
					}
					} /* switch */
				}

				if (w->subgroup) {
					Group *subgroup = (Group*) w->subgroup;
					group_write_html (f,dpy,subgroup);
				}

				w=w->next_in_line;
			}
			l=l->next;

			if (line_has_words)
				fprintf (f,"\n");
		}
	}
}

