
/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    table.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    11 Dec 2000
 * Purpose:        Table class;
 *                 drawing of tables and related graphics.
 *----------------------------------------------------------------------
 * Changes:
 * 11 Dec 2000, zeyr@excite.com: overhauled re-draw routine.
 *--------------------------------------------------------------------*/



#include <stdio.h>
#include <string.h>


#include "defs.h"
#include "malloc.h"
#include "main.h"
#include "word.h"
#include "error.h"
#include "table.h"
#include "group.h"


 
Table*
table_new () {
	Table *tbl;

	tbl = (Table*) my_malloc (sizeof(Table));
	if (!tbl) error_handler ("unable to allocate Table object");

	memset (tbl, 0, sizeof (Table));
	tbl->border_size=1;
	tbl->cell_padding=1;
	tbl->cell_spacing=1;
	return tbl;
}


void
table_free (Table *t) {
	CHECK_PARAM_NOT_NULL(t);

	my_free((char*) t);
}


void
table_cell_dump (FILE *f, CellInfo *c, char *name)
{
	if (!name) name="(none)";
	if (!c->is_leading_cell) {
		fprintf (f, "CELL '%s' -- not leading\n", name);
		return;
	}
	fprintf (f, "CELL '%s':\n",name);
	fprintf (f, "  optimal_width %ld\n  min_width %ld\n  optimal_height %ld\n", 
		c->optimal_width, 
		c->minimal_width, 
		c->optimal_height);
	fprintf (f, "  desired_width %ld\n", 
		c->desired_width);
	fprintf (f, "  colspan %ld\n  rowspan %ld\n", 
		c->col_span, c->row_span);
	fprintf (f, "END OF CELL\n");
}


void
table_dump (FILE *f, Table *t) {
	CellInfo *cell;
	int i,ix;

	if (!f) return;

	fprintf (f, "TABLE DUMP:\n");

	fprintf (f, "# rows = %d\n", t->rows);
	fprintf (f, "# min columns = %d\n", t->min_columns);
	fprintf (f, "# max columns = %d\n", t->columns);
	fprintf (f, "cell spacing = %d\n", t->cell_spacing);
	fprintf (f, "cell padding = %d\n", t->cell_padding);

	table_cell_dump (f,&t->overall,"entire-table");

	ix=i=0; 
	while (i < t->rows) {
		int j=0; 
		while (j < t->columns) {
			char str[100];
			sprintf (str, "r%d/c%d", i+1,j+1);
			table_cell_dump (f, &t->cells[ix++],str);
			j++;
		}
		i++;
	}

	fprintf (f, "LEADING CELL MATRIX:\n");

	ix=i=0; 
	while (i < t->rows) {
		int j=0; 
		while (j < t->columns) {
			if (t->cells[ix++].is_leading_cell)
				fprintf (f, "T ");
			else
				fprintf (f, "F ");
			j++;
		}
		fprintf (f, "\n");
		i++;
	}
	fprintf (f, "\n");

	fprintf (f, "END TABLE DUMP:\n");
}



void
table_redraw (DisplayDevice *dpy, Table *t, 
	Coordinate rx,
	Coordinate ry,
	Coordinate x0,
	Coordinate y0,
	Coordinate x1,
	Coordinate y1,
	Coordinate parents_x,
	Coordinate parents_y)
{
	CellInfo *cell;
	int i,j;
	int rows, max_columns;
	Coordinate x,y;
	Coordinate tx0, ty0;
	Coordinate tx1, ty1;
	Color top, left, right, bottom;
	Dimension border_size;
	Dimension cell_padding;
	Dimension cell_spacing;
	Coordinate base_x, base_y;
	Word *word;

	CHECK_PARAM_NOT_NULL(dpy);

	border_size = t->border_size;
	cell_padding = t->cell_padding;
	cell_spacing = t->cell_spacing;

	word = t->word;

	base_x = parents_x + word->x;
	base_y = parents_y + word->y;

	tx0 = base_x;
	ty0 = base_y;
	tx1 = tx0 + t->overall.optimal_width - 1;
	ty1 = ty0 + t->overall.optimal_height - 1;

	tx0 -= rx;
	tx1 -= rx;
	ty0 -= ry;
	ty1 -= ry;

	top = color_lookup ("grey7");
	left = color_lookup ("grey6");
	bottom = color_lookup ("grey2");
	right = color_lookup ("grey3");

	if (border_size) {
		dpy->draw_3d_rectangle (tx0, ty0, tx1, ty1,
			t->border_size,
			top,left,bottom,right);
	}

	dpy->reset_colors();

	i=0; 
	rows = t->rows;
	max_columns = t->columns;
	y = ty0 + border_size;

	for (i=0; i<rows; i++)
	{
		x = tx0 + border_size;

		for (j=0; j<max_columns; j++) 
		{
			CellInfo *cell;

			cell = &t->cells [max_columns * i + j];

			if (cell->is_leading_cell) {
				Group *subgroup;
				Word *td;
				Coordinate x2,y2;
				Dimension cumulative_width = 0;
				Dimension cumulative_height = 0;
				short rowspan, colspan;
				int k;

				td = cell->td;
				subgroup = (Group*) td->subgroup;

				if (cell_padding) {

					colspan = cell->col_span;
					rowspan = cell->row_span;

					for (k=0; k < colspan; k++) {
						cumulative_width += t->column_widths [k+j];
					}

					for (k=0; k < rowspan; k++) {
						cumulative_height += t->row_heights [k+i];
					}

					x2 = x + cumulative_width;
					x2 += colspan * 2 * cell_padding - 1;
					x2 += (colspan-1) * cell_spacing;
					y2 = y + cumulative_height;
					y2 += rowspan * 2 * cell_padding - 1;
					y2 += (rowspan-1) * cell_spacing;

					dpy->draw_3d_rectangle (x,y,x2,y2,
						cell_padding,
						bottom,right,top,left);

				}

				group_redraw_core (dpy, subgroup, rx, ry,
					x0,y0,x1,y1,base_x,base_y);
			}

			x += t->column_widths[j];
			x += cell_padding * 2;
			x += cell_spacing;
		}

		y += t->row_heights[i];
		y += cell_padding * 2;
		y += cell_spacing;
	}
}



/* Draw a table border that is a non three-dimensional gradiant.
 */
void
table_draw_border (DisplayDevice *dpy, 
	Coordinate x0, 
	Coordinate y0, 
	Coordinate x1,
	Coordinate y1, 
	Dimension size,
	Color color) 
{
	unsigned long r,g,b;
	unsigned long r2,g2,b2;

	r = (color >> 16) & 0xff;
	g = (color >> 8) & 0xff;
	b = color & 0xff;

	r *= 3; r /= 2;
	g *= 3; g /= 2;
	b *= 3; b /= 2;

	r2 = r/2;
	g2 = g/2;
	b2 = b/2;

	r2 = r - r2;
	g2 = g - g2;
	b2 = b - b2;

	r2 /= size;
	g2 /= size;
	b2 /= size;

	while (size--) {
		Color c = (r << 16) + (g << 8) + b;

		dpy->set_colors(c,c);

		dpy->draw_rectangle (x0,y0,x1,y1);
		x0++; y0++;
		x1--; y1--;

		r -= r2;
		g -= g2;
		b -= b2;
	}

	dpy->reset_colors();
}



#if 0
/*--------------------------------------------------------------------
 * Function:    table_redraw
 * Purpose:     Draws a table.
 * Note:        The Word param is the <table> word, not its children
 *------------------------------------------------------------------*/

/*
 * XX - rx,ry not used
 */

void
table_redraw (DisplayDevice *dpy, Table *tbl, Word *w, Coordinate rx, Coordinate ry)
{
	Word *w2;
	int i,row_num;
	Coordinate tx0, y;
	Coordinate tx1, ty1;
	Word **td_kids;

	CHECK_PARAM_NOT_NULL(dpy);
	CHECK_PARAM_NOT_NULL(tbl);
	CHECK_PARAM_NOT_NULL(w);

#if 0
	tx0 = tbl->x0;
	tx1 = tbl->x1;
	y =   tbl->y0;
	ty1 = tbl->y1;
#else
	tx0=w->x;
	tx1=w->x+w->width-1;
	y=w->y;
	ty1=w->y+w->ascent-1;
#endif

	td_kids = (Word**)  my_malloc (sizeof(Word*) * tbl->columns);

	table_draw_border (dpy,tx0-rx,y-ry,tx1-rx,ty1-ry,
	  tbl->border_size,tbl->border_color);

	/* Shrink our area to within the border.
	 */
	tx0 += tbl->border_size;
	y   += tbl->border_size;
	tx1 -= tbl->border_size;
	ty1 -= tbl->border_size;

	w=w->child;
	w2=w;
	row_num=0;
	while (w2) {
		if(w2->tag_number==TAG_TBODY) {
			w2=w2->child;
			continue;
		}

		if(w2->tag_number==TAG_TR) {
			Word *w3=w2->child;
			int num_columns=0;
			int row_ht;

			row_ht=tbl->row_heights[row_num++];

			/* Fetch table cells' contents
			 */
			while(w3) {
				if(w2->tag_number==TAG_TD) 
					td_kids[num_columns++] = w3->child;
				w3=w3->next;
			}

			/* Handle the case of just one cell per row. */
			if (num_columns == 0) {
				fprintf(stderr,"no columns in a table row. skipping...\n");
			} 
			else if (num_columns == 1) {
				/* assuming divider size is 1... XXX */
				dpy->draw_rectangle (tx0-1-rx, y-1-ry, tx1+1-rx, y+row_ht-ry);
				/* word_redraw_core (dpy, td_kids[0]); */
			} 
			/* ... versus multiple cells per row. */
			else if (num_columns > 1) {
				int x=tx0;
				for (i=0; i < num_columns; i++) 
				{
					Word *kid;
					dpy->draw_rectangle (x-1-rx, y-1-ry, 
						x+tbl->column_widths[i]-rx,
						y+row_ht-ry);

					kid = td_kids[i];
					if(kid) {
						/* word_redraw_core (dpy, kid); */
					} else {
						if (debug_mode)
							fprintf (stderr,"TD without child found (not a problem)\n");
					}

					x += tbl->column_widths[i]+1;
				}/*for*/
			}/*if*/
			y += (1+row_ht);
		}/*if*/
		w2=w2->next;
	}/*while*/
}

#endif
